import subprocess
import sys
from contextlib import contextmanager
from types import ModuleType, SimpleNamespace
from typing import Any
from unittest.mock import patch

import pytest

import brokk_code.__main__ as main_module
import brokk_code.git_utils as git_utils_module


def _stub_install_warmup(monkeypatch) -> None:
    monkeypatch.setattr(main_module, "ensure_jbang_ready", lambda: "/usr/local/bin/jbang")
    monkeypatch.setattr(main_module, "_run_install_prefetch", lambda _commands: None)
    monkeypatch.setattr(
        main_module,
        "wire_nvim_plugin_setup",
        lambda **_kwargs: SimpleNamespace(status="unsupported", path=None, detail=None),
    )


def test_main_version_subcommand_prints_version(monkeypatch, capsys) -> None:
    """Verify `brokk version` prints the package version and exits cleanly."""
    from brokk_code import __version__

    monkeypatch.setattr(sys, "argv", ["brokk", "version"])

    from brokk_code.__main__ import main

    main()

    captured = capsys.readouterr()
    assert f"brokk {__version__}" in captured.out


def test_main_defaults_to_tui(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}
    fake_app_module = ModuleType("brokk_code.app")

    class FakeApp:
        def __init__(self, **kwargs: Any):
            captured["kwargs"] = kwargs

        def run(self) -> None:
            captured["ran"] = True

    fake_app_module.BrokkApp = FakeApp
    monkeypatch.setitem(sys.modules, "brokk_code.app", fake_app_module)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "--workspace",
            str(tmp_path),
            "--session",
            "session-1",
            "--vendor",
            "OpenAI",
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["kwargs"]["workspace_dir"] == tmp_path.resolve()
    assert captured["kwargs"]["session_id"] == "session-1"
    assert captured["kwargs"]["vendor"] == "OpenAI"


def test_main_acp_routes_to_server(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {}
    fake_acp_module = ModuleType("brokk_code.acp_server")

    async def fake_run_acp_server(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs

    fake_acp_module.run_acp_server = fake_run_acp_server
    monkeypatch.setitem(sys.modules, "brokk_code.acp_server", fake_acp_module)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "acp",
            "--workspace",
            str(tmp_path),
            "--executor-stable",
            "--vendor",
            "Gemini",
        ],
    )

    main_module.main()

    assert captured["kwargs"]["workspace_dir"] == tmp_path.resolve()
    assert captured["kwargs"]["executor_snapshot"] is False
    assert captured["kwargs"]["vendor"] == "Gemini"


def test_main_mcp_routes_to_launcher(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {}
    jar_path = tmp_path / "brokk.jar"
    jar_path.write_text("dummy")

    def fake_run_mcp_server(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs

    monkeypatch.setattr(main_module, "run_mcp_server", fake_run_mcp_server)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "mcp",
            "--workspace",
            str(tmp_path),
            "--jar",
            str(jar_path),
            "--executor-version",
            "0.99.0",
        ],
    )

    main_module.main()

    assert captured["kwargs"]["workspace_dir"] == tmp_path.resolve()
    assert captured["kwargs"]["jar_path"] == jar_path.resolve()
    assert captured["kwargs"]["executor_version"] == "0.99.0"
    # Ensure no residual ide parameter is passed from CLI to run_acp_server
    assert "ide" not in captured["kwargs"]


def test_main_acp_accepts_legacy_ide_flag_but_ignores_it(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {}
    fake_acp_module = ModuleType("brokk_code.acp_server")

    async def fake_run_acp_server(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs

    fake_acp_module.run_acp_server = fake_run_acp_server
    monkeypatch.setitem(sys.modules, "brokk_code.acp_server", fake_acp_module)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "acp",
            "--workspace",
            str(tmp_path),
            "--executor-stable",
            "--vendor",
            "Gemini",
            "--ide",
            "zed",
        ],
    )

    main_module.main()

    # Still routes correctly
    assert captured["kwargs"]["workspace_dir"] == tmp_path.resolve()
    assert captured["kwargs"]["executor_snapshot"] is False
    assert captured["kwargs"]["vendor"] == "Gemini"
    # Critically: ide is not forwarded to run_acp_server
    assert "ide" not in captured["kwargs"]


def test_main_acp_rejects_extra_positional(monkeypatch, tmp_path) -> None:
    monkeypatch.setattr(
        sys,
        "argv",
        ["brokk", "acp", "zed", "--workspace", str(tmp_path)],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 2


def test_main_install_zed_routes_to_installer(monkeypatch, tmp_path, capsys) -> None:
    captured: dict[str, Any] = {}

    def fake_configure_zed_acp_settings(
        *, force: bool = False, settings_path=None, uvx_command=None
    ):
        captured["force"] = force
        return tmp_path / ".config" / "zed" / "settings.json"

    monkeypatch.setattr(main_module, "configure_zed_acp_settings", fake_configure_zed_acp_settings)
    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(sys, "argv", ["brokk", "install", "zed", "--force"])

    main_module.main()

    output = capsys.readouterr().out
    assert captured["force"] is True
    assert "Configured Zed ACP integration" in output


def test_main_install_zed_conflict_exits_nonzero(monkeypatch) -> None:
    def fake_configure_zed_acp_settings(
        *, force: bool = False, settings_path=None, uvx_command=None
    ):
        raise main_module.ExistingBrokkCodeEntryError("exists")

    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(main_module, "configure_zed_acp_settings", fake_configure_zed_acp_settings)
    monkeypatch.setattr(sys, "argv", ["brokk", "install", "zed"])

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 1


def test_main_install_zed_invalid_json_exits_nonzero(monkeypatch) -> None:
    def fake_configure_zed_acp_settings(
        *, force: bool = False, settings_path=None, uvx_command=None
    ):
        raise ValueError("Could not parse as JSON/JSONC")

    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(main_module, "configure_zed_acp_settings", fake_configure_zed_acp_settings)
    monkeypatch.setattr(sys, "argv", ["brokk", "install", "zed"])

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 1


def test_main_install_intellij_routes_to_installer(monkeypatch, tmp_path, capsys) -> None:
    captured: dict[str, Any] = {}

    def fake_configure_intellij_acp_settings(
        *, force: bool = False, settings_path=None, uvx_command=None
    ):
        captured["force"] = force
        return tmp_path / "intellij-config"

    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(
        main_module, "configure_intellij_acp_settings", fake_configure_intellij_acp_settings
    )
    monkeypatch.setattr(sys, "argv", ["brokk", "install", "intellij", "--force"])

    main_module.main()

    output = capsys.readouterr().out
    assert captured["force"] is True
    assert "Configured IntelliJ ACP integration" in output


def test_main_install_nvim_routes_to_installer(monkeypatch, tmp_path, capsys) -> None:
    captured: dict[str, Any] = {}

    def fake_configure_nvim_codecompanion_acp_settings(*, force: bool = False, settings_path=None):
        captured["force"] = force
        return tmp_path / ".config" / "nvim" / "lua" / "brokk" / "brokk_codecompanion.lua"

    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(
        main_module,
        "configure_nvim_codecompanion_acp_settings",
        fake_configure_nvim_codecompanion_acp_settings,
    )
    monkeypatch.setattr(
        sys, "argv", ["brokk", "install", "nvim", "--plugin", "codecompanion", "--force"]
    )

    main_module.main()

    output = capsys.readouterr().out
    assert captured["force"] is True
    assert "Configured Neovim CodeCompanion ACP adapter" in output


def test_main_install_nvim_conflict_exits_nonzero(monkeypatch) -> None:
    def fake_configure_nvim_codecompanion_acp_settings(*, force: bool = False, settings_path=None):
        raise main_module.ExistingBrokkCodeEntryError("exists")

    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(
        main_module,
        "configure_nvim_codecompanion_acp_settings",
        fake_configure_nvim_codecompanion_acp_settings,
    )
    monkeypatch.setattr(sys, "argv", ["brokk", "install", "nvim", "--plugin", "codecompanion"])

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 1


def test_main_install_neovim_with_plugin_codecompanion_routes_to_installer(
    monkeypatch, tmp_path, capsys
) -> None:
    captured: dict[str, Any] = {}

    def fake_configure_nvim_codecompanion_acp_settings(*, force: bool = False, settings_path=None):
        captured["force"] = force
        return tmp_path / ".config" / "nvim" / "lua" / "brokk" / "brokk_codecompanion.lua"

    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(
        main_module,
        "configure_nvim_codecompanion_acp_settings",
        fake_configure_nvim_codecompanion_acp_settings,
    )
    monkeypatch.setattr(
        sys, "argv", ["brokk", "install", "neovim", "--plugin", "codecompanion", "--force"]
    )

    main_module.main()

    output = capsys.readouterr().out
    assert captured["force"] is True
    assert "Configured Neovim CodeCompanion ACP adapter" in output


def test_main_install_neovim_with_plugin_avante_routes_to_installer(
    monkeypatch, tmp_path, capsys
) -> None:
    captured: dict[str, Any] = {}

    def fake_configure_nvim_avante_acp_settings(*, force: bool = False, settings_path=None):
        captured["force"] = force
        return tmp_path / ".config" / "nvim" / "lua" / "brokk" / "brokk_avante.lua"

    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(
        main_module,
        "configure_nvim_avante_acp_settings",
        fake_configure_nvim_avante_acp_settings,
    )
    monkeypatch.setattr(
        sys, "argv", ["brokk", "install", "neovim", "--plugin", "avante", "--force"]
    )

    main_module.main()

    output = capsys.readouterr().out
    assert captured["force"] is True
    assert "Configured Neovim Avante ACP provider" in output


def test_main_install_plugin_with_non_neovim_target_exits_nonzero(monkeypatch) -> None:
    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(
        sys,
        "argv",
        ["brokk", "install", "zed", "--plugin", "avante"],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 1


def test_main_install_neovim_routes_to_installer(monkeypatch, tmp_path, capsys) -> None:
    captured: dict[str, Any] = {}

    def fake_configure_nvim_codecompanion_acp_settings(*, force: bool = False, settings_path=None):
        captured["force"] = force
        return tmp_path / ".config" / "nvim" / "lua" / "brokk" / "brokk_codecompanion.lua"

    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(
        main_module,
        "configure_nvim_codecompanion_acp_settings",
        fake_configure_nvim_codecompanion_acp_settings,
    )
    monkeypatch.setattr(sys, "argv", ["brokk", "install", "neovim", "--force"])

    main_module.main()

    output = capsys.readouterr().out
    assert captured["force"] is True
    assert "Configured Neovim CodeCompanion ACP adapter" in output


def test_main_install_neovim_with_plugin_codecompanion_conflict_exits_nonzero(monkeypatch) -> None:
    def fake_configure_nvim_codecompanion_acp_settings(*, force: bool = False, settings_path=None):
        raise main_module.ExistingBrokkCodeEntryError("exists")

    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(
        main_module,
        "configure_nvim_codecompanion_acp_settings",
        fake_configure_nvim_codecompanion_acp_settings,
    )
    monkeypatch.setattr(sys, "argv", ["brokk", "install", "neovim", "--plugin", "codecompanion"])

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 1


def test_main_install_verbose_prints_prefetch_command(monkeypatch, tmp_path, capsys) -> None:
    monkeypatch.setattr(main_module, "resolve_jbang_binary", lambda: None)
    prefetch_invoked: dict[str, bool] = {"called": False}

    def fake_run_install_prefetch(_commands: list[tuple[str, list[str]]]) -> None:
        prefetch_invoked["called"] = True

    monkeypatch.setattr(main_module, "_run_install_prefetch", fake_run_install_prefetch)

    def fake_configure_zed_acp_settings(
        *, force: bool = False, settings_path=None, uvx_command=None
    ):
        return tmp_path / ".config" / "zed" / "settings.json"

    monkeypatch.setattr(main_module, "configure_zed_acp_settings", fake_configure_zed_acp_settings)
    monkeypatch.setattr(sys, "argv", ["brokk", "install", "zed", "-v"])

    main_module.main()

    output = capsys.readouterr().out.strip().splitlines()
    assert any("Configured Zed ACP integration" in line for line in output)
    assert any("jbang" in line for line in output)
    assert any("--main" in line for line in output)
    assert prefetch_invoked["called"] is False


def test_main_install_intellij_conflict_exits_nonzero(monkeypatch) -> None:
    def fake_configure_intellij_acp_settings(
        *, force: bool = False, settings_path=None, uvx_command=None
    ):
        raise main_module.ExistingBrokkCodeEntryError("exists")

    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(
        main_module, "configure_intellij_acp_settings", fake_configure_intellij_acp_settings
    )
    monkeypatch.setattr(sys, "argv", ["brokk", "install", "intellij"])

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 1


def test_main_install_intellij_invalid_json_exits_nonzero(monkeypatch) -> None:
    def fake_configure_intellij_acp_settings(
        *, force: bool = False, settings_path=None, uvx_command=None
    ):
        raise ValueError("Could not parse as JSON")

    _stub_install_warmup(monkeypatch)
    monkeypatch.setattr(
        main_module, "configure_intellij_acp_settings", fake_configure_intellij_acp_settings
    )
    monkeypatch.setattr(sys, "argv", ["brokk", "install", "intellij"])

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 1


def test_main_install_mcp_routes_to_installer(monkeypatch, tmp_path, capsys) -> None:
    captured: dict[str, Any] = {}
    prefetched: dict[str, Any] = {}

    def fake_configure_claude_code_mcp_settings(
        *, force: bool = False, settings_path=None, jbang_path=None, uvx_command=None
    ):
        captured["claude_force"] = force
        return tmp_path / "claude.json"

    def fake_configure_codex_mcp_settings(
        *, force: bool = False, settings_path=None, jbang_path=None, uvx_command=None
    ):
        captured["codex_force"] = force
        return tmp_path / "codex.toml"

    def fake_run_install_prefetch(commands):
        prefetched["commands"] = commands

    def fake_install_codex_mcp_workspace_skill(*, skills_path=None):
        return tmp_path / ".codex" / "skills" / "brokk-mcp-workspace" / "SKILL.md"

    monkeypatch.setattr(
        main_module,
        "configure_claude_code_mcp_settings",
        fake_configure_claude_code_mcp_settings,
    )
    monkeypatch.setattr(
        main_module,
        "configure_codex_mcp_settings",
        fake_configure_codex_mcp_settings,
    )
    monkeypatch.setattr(
        main_module,
        "install_codex_mcp_workspace_skill",
        fake_install_codex_mcp_workspace_skill,
    )
    monkeypatch.setattr(main_module, "ensure_jbang_ready", lambda: "/usr/local/bin/jbang")
    monkeypatch.setattr(main_module, "_run_install_prefetch", fake_run_install_prefetch)
    monkeypatch.setattr(
        sys,
        "argv",
        ["brokk", "install", "mcp", "--force"],
    )

    main_module.main()

    output = capsys.readouterr().out
    assert captured["claude_force"] is True
    assert captured["codex_force"] is True
    assert "Configured Claude Code MCP integration" in output
    assert "Configured Codex MCP integration" in output
    assert "Installed Codex MCP workspace skill" in output
    assert "MCP runtime" in str(prefetched["commands"][0][0])


def test_main_uses_git_repo_root_for_nested_workspace(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}
    fake_app_module = ModuleType("brokk_code.app")
    repo_root = tmp_path / "repo"
    nested_workspace = repo_root / "src" / "feature"
    nested_workspace.mkdir(parents=True)
    (repo_root / ".git").mkdir()

    class FakeApp:
        def __init__(self, **kwargs: Any):
            captured["kwargs"] = kwargs

        def run(self) -> None:
            captured["ran"] = True

    fake_app_module.BrokkApp = FakeApp
    monkeypatch.setitem(sys.modules, "brokk_code.app", fake_app_module)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "--workspace",
            str(nested_workspace),
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["kwargs"]["workspace_dir"] == repo_root.resolve()


def test_main_keeps_workspace_when_not_in_git_repo(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}
    fake_app_module = ModuleType("brokk_code.app")
    nested_workspace = tmp_path / "workspace" / "src"
    nested_workspace.mkdir(parents=True)

    class FakeApp:
        def __init__(self, **kwargs: Any):
            captured["kwargs"] = kwargs

        def run(self) -> None:
            captured["ran"] = True

    fake_app_module.BrokkApp = FakeApp
    monkeypatch.setitem(sys.modules, "brokk_code.app", fake_app_module)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "--workspace",
            str(nested_workspace),
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["kwargs"]["workspace_dir"] == nested_workspace.resolve()


def test_main_resume_routes_correctly(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}
    fake_app_module = ModuleType("brokk_code.app")

    class FakeApp:
        def __init__(self, **kwargs: Any):
            captured["kwargs"] = kwargs

        def run(self) -> None:
            captured["ran"] = True

    fake_app_module.BrokkApp = FakeApp
    monkeypatch.setitem(sys.modules, "brokk_code.app", fake_app_module)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "resume",
            "session-xyz",
            "--workspace",
            str(tmp_path),
            "--vendor",
            "Anthropic",
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["kwargs"]["session_id"] == "session-xyz"
    assert captured["kwargs"]["resume_session"] is False
    assert captured["kwargs"]["workspace_dir"] == tmp_path.resolve()
    assert captured["kwargs"]["vendor"] == "Anthropic"


def test_main_sessions_routes_correctly(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}
    fake_app_module = ModuleType("brokk_code.app")

    class FakeApp:
        def __init__(self, **kwargs: Any):
            captured["kwargs"] = kwargs

        def run(self) -> None:
            captured["ran"] = True

    fake_app_module.BrokkApp = FakeApp
    monkeypatch.setitem(sys.modules, "brokk_code.app", fake_app_module)

    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "sessions",
            "--workspace",
            str(tmp_path),
            "--vendor",
            "OpenAI",
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    # pick_session must be True for the sessions command
    assert captured["kwargs"]["pick_session"] is True
    # workspace_dir should be resolved
    assert captured["kwargs"]["workspace_dir"] == tmp_path.resolve()
    # vendor should be passed through
    assert captured["kwargs"]["vendor"] == "OpenAI"
    # sessions command overrides session_id and resume_session
    assert captured["kwargs"]["session_id"] is None
    assert captured["kwargs"]["resume_session"] is False


def test_main_sessions_rejects_positional_args(monkeypatch, tmp_path) -> None:
    monkeypatch.setattr(
        sys,
        "argv",
        ["brokk", "sessions", "unexpected-arg", "--workspace", str(tmp_path)],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 2


def test_main_issue_create_routes_correctly(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}
    temp_workspace = tmp_path / "temp-create"
    temp_workspace.mkdir()

    async def fake_run_headless_job(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    @contextmanager
    def fake_temp_checkout(**kwargs: Any):
        captured["checkout_kwargs"] = kwargs
        yield temp_workspace

    monkeypatch.setattr(main_module, "run_headless_job", fake_run_headless_job)
    monkeypatch.setattr(main_module, "_temporary_issue_repo_checkout", fake_temp_checkout)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "create",
            "Broken build",
            "--workspace",
            str(tmp_path),
            "--github-token",
            "ghp_123",
            "--repo-owner",
            "acme",
            "--repo-name",
            "tools",
            "--planner-model",
            "custom-model",
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["checkout_kwargs"]["repo_owner"] == "acme"
    assert captured["checkout_kwargs"]["repo_name"] == "tools"
    assert captured["checkout_kwargs"]["github_token"] == "ghp_123"
    assert captured["checkout_kwargs"]["action_label"] == "Issue create"
    assert captured["kwargs"]["workspace_dir"] == temp_workspace
    assert captured["kwargs"]["task_input"] == "Broken build"
    assert captured["kwargs"]["mode"] == "ISSUE_WRITER"
    assert captured["kwargs"]["planner_model"] == "custom-model"
    assert captured["kwargs"]["tags"]["github_token"] == "ghp_123"
    assert captured["kwargs"]["tags"]["repo_owner"] == "acme"
    assert captured["kwargs"]["tags"]["repo_name"] == "tools"


def test_main_issue_create_missing_prompt_exits_nonzero(monkeypatch, tmp_path) -> None:
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "create",
            "--github-token",
            "ghp_123",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_issue_create_validation_missing_token(monkeypatch, capsys) -> None:
    monkeypatch.setattr(
        sys,
        "argv",
        ["brokk", "issue", "create", "test", "--repo-owner", "o", "--repo-name", "r"],
    )
    # Ensure no env var leaks in
    monkeypatch.delenv("GITHUB_TOKEN", raising=False)

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 1
    assert "Error: --github-token is required for issue create" in capsys.readouterr().err


def test_main_issue_solve_validation_invalid_owner(monkeypatch, capsys) -> None:
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "solve",
            "--issue-number",
            "1",
            "--github-token",
            "t",
            "--repo-owner",
            "invalid/owner",
            "--repo-name",
            "r",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 1
    err = capsys.readouterr().err
    assert "Error: Invalid --repo-owner 'invalid/owner'" in err
    assert "^[A-Za-z0-9_.-]+$" in err


def test_main_issue_create_respects_env_github_token(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}
    temp_workspace = tmp_path / "temp-create-env"
    temp_workspace.mkdir()

    async def fake_run_headless_job(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    @contextmanager
    def fake_temp_checkout(**kwargs: Any):
        captured["checkout_kwargs"] = kwargs
        yield temp_workspace

    monkeypatch.setattr(main_module, "run_headless_job", fake_run_headless_job)
    monkeypatch.setattr(main_module, "_temporary_issue_repo_checkout", fake_temp_checkout)
    monkeypatch.setenv("GITHUB_TOKEN", "env-token")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "create",
            "Broken build",
            "--repo-owner",
            "acme",
            "--repo-name",
            "tools",
        ],
    )

    main_module.main()

    assert captured["kwargs"]["tags"]["github_token"] == "env-token"
    assert captured["checkout_kwargs"]["github_token"] == "env-token"
    assert captured["kwargs"]["workspace_dir"] == temp_workspace
    assert captured["kwargs"]["planner_model"] == "gemini-3-flash-preview"
    assert captured["kwargs"]["planner_reasoning_level"] == "disable"
    assert captured["kwargs"]["verbose"] is False


def test_main_issue_create_verbose_routes_correctly(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}
    temp_workspace = tmp_path / "temp-create-verbose"
    temp_workspace.mkdir()

    async def fake_run_headless_job(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    @contextmanager
    def fake_temp_checkout(**kwargs: Any):
        captured["checkout_kwargs"] = kwargs
        yield temp_workspace

    monkeypatch.setattr(main_module, "run_headless_job", fake_run_headless_job)
    monkeypatch.setattr(main_module, "_temporary_issue_repo_checkout", fake_temp_checkout)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "create",
            "Broken build",
            "-v",
            "--github-token",
            "ghp_verbose",
            "--repo-owner",
            "acme",
            "--repo-name",
            "tools",
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["checkout_kwargs"]["action_label"] == "Issue create"
    assert captured["kwargs"]["workspace_dir"] == temp_workspace
    assert captured["kwargs"]["verbose"] is True


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_headless_job_creates_session_before_wait_ready(
    mock_executor_class, tmp_path
) -> None:
    """Verifies that run_headless_job creates a session before polling for readiness.

    This ordering is required because the Java executor's /health/ready endpoint
    only returns 200 OK after a session has been created.
    """
    from unittest.mock import AsyncMock

    call_order: list[str] = []
    mock_manager = mock_executor_class.return_value

    async def mock_start():
        call_order.append("start")

    async def mock_create_session(name: str = ""):
        call_order.append("create_session")
        return "session-123"

    async def mock_wait_ready(timeout: float = 30.0):
        call_order.append("wait_ready")
        return True

    async def mock_submit_job(**kwargs):
        call_order.append("submit_job")
        return "job-456"

    async def mock_stream_events(job_id: str):
        # Yield a terminal state event to end the job
        yield {"type": "STATE_CHANGE", "state": "COMPLETED"}

    async def mock_stop():
        call_order.append("stop")

    mock_manager.start = AsyncMock(side_effect=mock_start)
    mock_manager.create_session = AsyncMock(side_effect=mock_create_session)
    mock_manager.wait_ready = AsyncMock(side_effect=mock_wait_ready)
    mock_manager.submit_job = AsyncMock(side_effect=mock_submit_job)
    mock_manager.stream_events = mock_stream_events
    mock_manager.stop = AsyncMock(side_effect=mock_stop)

    await main_module.run_headless_job(
        workspace_dir=tmp_path,
        task_input="Test task",
        planner_model="test-model",
        mode="LUTZ",
        tags={},
    )

    # Verify the critical ordering: create_session MUST come before wait_ready
    assert "start" in call_order
    assert "create_session" in call_order
    assert "wait_ready" in call_order
    assert "submit_job" in call_order

    start_idx = call_order.index("start")
    create_session_idx = call_order.index("create_session")
    wait_ready_idx = call_order.index("wait_ready")
    submit_job_idx = call_order.index("submit_job")

    # The critical assertion: session must be created before waiting for readiness
    assert start_idx < create_session_idx, "start() must be called before create_session()"
    assert create_session_idx < wait_ready_idx, (
        "create_session() must be called before wait_ready()"
    )
    assert wait_ready_idx < submit_job_idx, "wait_ready() must be called before submit_job()"


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_headless_job_reports_failed_terminal_state(
    mock_executor_class, tmp_path, capsys
) -> None:
    from unittest.mock import AsyncMock

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.submit_job = AsyncMock(return_value="job-456")

    async def mock_stream_events(job_id: str):
        assert job_id == "job-456"
        yield {"type": "STATE_CHANGE", "state": "RUNNING"}
        yield {"type": "ERROR", "message": "GitHub API returned 403"}
        yield {"type": "STATE_CHANGE", "state": "FAILED"}

    mock_manager.stream_events = mock_stream_events
    mock_manager.stop = AsyncMock()

    with pytest.raises(SystemExit) as exc:
        await main_module.run_headless_job(
            workspace_dir=tmp_path,
            task_input="Create issue",
            planner_model="test-model",
            mode="ISSUE_WRITER",
            tags={},
        )

    captured = capsys.readouterr()
    assert exc.value.code == 1
    assert "Error event: GitHub API returned 403" in captured.err
    assert "ISSUE_WRITER job ended with state FAILED." in captured.err
    assert "Last error: GitHub API returned 403" in captured.err
    mock_manager.stop.assert_awaited_once()


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_headless_job_reports_stage_on_submit_failure(
    mock_executor_class, tmp_path, capsys
) -> None:
    from unittest.mock import AsyncMock

    from brokk_code.executor import ExecutorError

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.submit_job = AsyncMock(side_effect=ExecutorError("401 Unauthorized"))
    mock_manager.stop = AsyncMock()

    with pytest.raises(SystemExit) as exc:
        await main_module.run_headless_job(
            workspace_dir=tmp_path,
            task_input="Create issue",
            planner_model="test-model",
            mode="ISSUE_WRITER",
            tags={},
        )

    captured = capsys.readouterr()
    assert exc.value.code == 1
    assert (
        "Executor error during ISSUE_WRITER job (submitting job): 401 Unauthorized" in captured.err
    )
    mock_manager.stop.assert_awaited_once()


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_headless_job_uses_nested_event_data_for_errors_and_quiet_notifications(
    mock_executor_class, tmp_path, capsys
) -> None:
    from unittest.mock import AsyncMock

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.submit_job = AsyncMock(return_value="job-456")

    async def mock_stream_events(job_id: str):
        assert job_id == "job-456"
        yield {"type": "NOTIFICATION", "data": {"level": "INFO", "message": ""}}
        yield {"type": "NOTIFICATION", "data": {"level": "WARN", "message": "rate limit near"}}
        yield {"type": "LLM_TOKEN", "data": {"token": "hello"}}
        yield {"type": "ERROR", "data": {"message": "executor boom"}}
        yield {"type": "STATE_CHANGE", "data": {"state": "FAILED"}}

    mock_manager.stream_events = mock_stream_events
    mock_manager.stop = AsyncMock()

    with pytest.raises(SystemExit) as exc:
        await main_module.run_headless_job(
            workspace_dir=tmp_path,
            task_input="Create issue",
            planner_model="test-model",
            mode="ISSUE_WRITER",
            tags={},
        )

    captured = capsys.readouterr()
    assert exc.value.code == 1
    assert "[INFO] None" not in captured.out
    assert "[WARN] rate limit near" in captured.out
    assert "hello" not in captured.out
    assert "Error event: executor boom" in captured.err
    assert "Unknown error event" not in captured.err


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_headless_job_verbose_shows_full_event_output(
    mock_executor_class, tmp_path, capsys
) -> None:
    from unittest.mock import AsyncMock

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.submit_job = AsyncMock(return_value="job-456")

    async def mock_stream_events(job_id: str):
        assert job_id == "job-456"
        yield {"type": "NOTIFICATION", "data": {"level": "INFO", "message": "planning"}}
        yield {"type": "STATE_CHANGE", "data": {"state": "RUNNING"}}
        yield {"type": "LLM_TOKEN", "data": {"token": "hello"}}
        yield {"type": "COMMAND_RESULT", "data": {"command": "gh issue create", "output": "ok"}}
        yield {"type": "TOOL_OUTPUT", "data": {"text": "tool text"}}
        yield {"type": "STATE_CHANGE", "data": {"state": "COMPLETED"}}

    mock_manager.stream_events = mock_stream_events
    mock_manager.stop = AsyncMock()

    await main_module.run_headless_job(
        workspace_dir=tmp_path,
        task_input="Create issue",
        planner_model="test-model",
        mode="ISSUE_WRITER",
        tags={},
        verbose=True,
    )

    captured = capsys.readouterr()
    assert "[INFO] planning" in captured.out
    assert "Job state: RUNNING" in captured.out
    assert "hello" in captured.out
    assert "[COMMAND_RESULT]" in captured.out
    assert "[TOOL_OUTPUT]" in captured.out


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_headless_job_exits_nonzero_on_error_event_without_failed_state(
    mock_executor_class, tmp_path, capsys
) -> None:
    from unittest.mock import AsyncMock

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.submit_job = AsyncMock(return_value="job-456")

    async def mock_stream_events(job_id: str):
        assert job_id == "job-456"
        yield {"type": "NOTIFICATION", "data": {"level": "INFO", "message": "Job started"}}
        yield {"type": "ERROR", "data": {"message": "parseIssueResponse: invalid JSON"}}
        # Stream ends without a terminal FAILED/CANCELLED state event.

    mock_manager.stream_events = mock_stream_events
    mock_manager.stop = AsyncMock()

    with pytest.raises(SystemExit) as exc:
        await main_module.run_headless_job(
            workspace_dir=tmp_path,
            task_input="Create issue",
            planner_model="test-model",
            mode="ISSUE_WRITER",
            tags={},
        )

    captured = capsys.readouterr()
    assert exc.value.code == 1
    assert "Job finished." not in captured.out
    assert "ISSUE_WRITER job ended with errors (last observed state: UNKNOWN)." in captured.err
    assert "Last error: parseIssueResponse: invalid JSON" in captured.err


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_headless_job_prints_issue_created_link_from_suppressed_tokens(
    mock_executor_class, tmp_path, capsys
) -> None:
    from unittest.mock import AsyncMock

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.submit_job = AsyncMock(return_value="job-456")

    async def mock_stream_events(job_id: str):
        assert job_id == "job-456"
        yield {"type": "LLM_TOKEN", "data": {"token": "Created issue: "}}
        yield {
            "type": "LLM_TOKEN",
            "data": {"token": "https://github.com/brokkai/brokk/issues/123"},
        }
        yield {"type": "STATE_CHANGE", "data": {"state": "COMPLETED"}}

    mock_manager.stream_events = mock_stream_events
    mock_manager.stop = AsyncMock()

    await main_module.run_headless_job(
        workspace_dir=tmp_path,
        task_input="Create issue",
        planner_model="test-model",
        mode="ISSUE_WRITER",
        tags={},
    )

    captured = capsys.readouterr()
    assert "Issue created: https://github.com/brokkai/brokk/issues/123" in captured.out
    assert "Job submitted:" not in captured.out
    assert "Job finished." not in captured.out


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_headless_job_prints_issue_created_link_from_issue_writer_notification(
    mock_executor_class, tmp_path, capsys
) -> None:
    from unittest.mock import AsyncMock

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.submit_job = AsyncMock(return_value="job-456")

    async def mock_stream_events(job_id: str):
        assert job_id == "job-456"
        yield {
            "type": "NOTIFICATION",
            "data": {
                "level": "INFO",
                "message": "ISSUE_WRITER: issue created I_kwDOXYZ https://github.com/brokkai/brokk/issues/456",
            },
        }
        yield {"type": "STATE_CHANGE", "data": {"state": "COMPLETED"}}

    mock_manager.stream_events = mock_stream_events
    mock_manager.stop = AsyncMock()

    await main_module.run_headless_job(
        workspace_dir=tmp_path,
        task_input="Create issue",
        planner_model="test-model",
        mode="ISSUE_WRITER",
        tags={},
    )

    captured = capsys.readouterr()
    assert "Issue created: https://github.com/brokkai/brokk/issues/456" in captured.out


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_headless_job_prints_issue_created_link_from_tool_output_result_text(
    mock_executor_class, tmp_path, capsys
) -> None:
    from unittest.mock import AsyncMock

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.submit_job = AsyncMock(return_value="job-456")

    async def mock_stream_events(job_id: str):
        assert job_id == "job-456"
        yield {
            "type": "TOOL_OUTPUT",
            "data": {
                "resultText": "Created: https://github.com/brokkai/brokk/issues/789",
                "name": "createGitHubIssue",
                "status": "SUCCESS",
            },
        }
        yield {"type": "STATE_CHANGE", "data": {"state": "COMPLETED"}}

    mock_manager.stream_events = mock_stream_events
    mock_manager.stop = AsyncMock()

    await main_module.run_headless_job(
        workspace_dir=tmp_path,
        task_input="Create issue",
        planner_model="test-model",
        mode="ISSUE_WRITER",
        tags={},
    )

    captured = capsys.readouterr()
    assert "Issue created: https://github.com/brokkai/brokk/issues/789" in captured.out


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_headless_job_prints_issue_created_link_from_structured_issue_created_event(
    mock_executor_class, tmp_path, capsys
) -> None:
    from unittest.mock import AsyncMock

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.submit_job = AsyncMock(return_value="job-456")

    async def mock_stream_events(job_id: str):
        assert job_id == "job-456"
        yield {
            "type": "ISSUE_CREATED",
            "data": {
                "issueId": "#987",
                "issueUrl": "https://github.com/brokkai/brokk/issues/987",
                "repoOwner": "brokkai",
                "repoName": "brokk",
            },
        }
        yield {"type": "STATE_CHANGE", "data": {"state": "COMPLETED"}}

    mock_manager.stream_events = mock_stream_events
    mock_manager.stop = AsyncMock()

    await main_module.run_headless_job(
        workspace_dir=tmp_path,
        task_input="Create issue",
        planner_model="test-model",
        mode="ISSUE_WRITER",
        tags={},
    )

    captured = capsys.readouterr()
    assert "Issue created: https://github.com/brokkai/brokk/issues/987" in captured.out


def test_main_issue_solve_routes_correctly(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}
    temp_workspace = tmp_path / "temp-copy"
    temp_workspace.mkdir()

    async def fake_run_headless_job(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    @contextmanager
    def fake_temp_workspace(**kwargs: Any):
        captured["temp_workspace_input"] = kwargs
        yield temp_workspace

    monkeypatch.setattr(main_module, "run_headless_job", fake_run_headless_job)
    monkeypatch.setattr(main_module, "_temporary_issue_repo_checkout", fake_temp_workspace)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "solve",
            "--issue-number",
            "123",
            "--workspace",
            str(tmp_path),
            "--github-token",
            "ghp_solve",
            "--repo-owner",
            "acme",
            "--repo-name",
            "tools",
            "--skip-verification",
            "--max-issue-fix-attempts",
            "7",
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["temp_workspace_input"]["repo_owner"] == "acme"
    assert captured["temp_workspace_input"]["repo_name"] == "tools"
    assert captured["temp_workspace_input"]["github_token"] == "ghp_solve"
    assert captured["temp_workspace_input"]["action_label"] == "Issue solve"
    assert captured["kwargs"]["mode"] == "ISSUE"
    assert captured["kwargs"]["workspace_dir"] == temp_workspace
    assert captured["kwargs"]["task_input"] == "Resolve GitHub Issue #123"
    assert captured["kwargs"]["tags"]["issue_number"] == "123"
    assert captured["kwargs"]["tags"]["github_token"] == "ghp_solve"
    assert captured["kwargs"]["skip_verification"] is True
    assert captured["kwargs"]["max_issue_fix_attempts"] == 7


def test_main_issue_solve_temp_workspace_cleanup_on_keyboard_interrupt(
    monkeypatch, tmp_path
) -> None:
    captured: dict[str, Any] = {"cleaned": False}
    temp_workspace = tmp_path / "temp-copy"
    temp_workspace.mkdir()

    async def fake_run_headless_job(**kwargs: Any) -> None:
        raise KeyboardInterrupt

    @contextmanager
    def fake_temp_workspace(**kwargs: Any):
        captured["temp_workspace_input"] = kwargs
        try:
            yield temp_workspace
        finally:
            captured["cleaned"] = True

    monkeypatch.setattr(main_module, "run_headless_job", fake_run_headless_job)
    monkeypatch.setattr(main_module, "_temporary_issue_repo_checkout", fake_temp_workspace)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "solve",
            "--issue-number",
            "123",
            "--workspace",
            str(tmp_path),
            "--github-token",
            "ghp_solve",
            "--repo-owner",
            "acme",
            "--repo-name",
            "tools",
        ],
    )

    with pytest.raises(KeyboardInterrupt):
        main_module.main()

    assert captured["temp_workspace_input"]["repo_owner"] == "acme"
    assert captured["cleaned"] is True


def test_main_issue_solve_missing_number_exits_nonzero(monkeypatch) -> None:
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "solve",
            "--repo-owner",
            "acme",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_issue_solve_missing_github_token_exits_nonzero(monkeypatch) -> None:
    monkeypatch.delenv("GITHUB_TOKEN", raising=False)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "solve",
            "--issue-number",
            "123",
            "--repo-owner",
            "acme",
            "--repo-name",
            "tools",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_issue_solve_missing_repo_owner_exits_nonzero(monkeypatch) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "token")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "solve",
            "--issue-number",
            "123",
            "--repo-name",
            "tools",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_issue_solve_missing_repo_name_exits_nonzero(monkeypatch) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "token")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "solve",
            "--issue-number",
            "123",
            "--repo-owner",
            "acme",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_issue_solve_invalid_repo_owner_exits_nonzero(monkeypatch) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "token")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "solve",
            "--issue-number",
            "123",
            "--repo-owner",
            "invalid/owner",
            "--repo-name",
            "tools",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_issue_solve_invalid_repo_name_exits_nonzero(monkeypatch) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "token")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "solve",
            "--issue-number",
            "123",
            "--repo-owner",
            "acme",
            "--repo-name",
            "invalid/repo",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_issue_create_missing_github_token_exits_nonzero(monkeypatch) -> None:
    monkeypatch.delenv("GITHUB_TOKEN", raising=False)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "create",
            "new issue",
            "--repo-owner",
            "acme",
            "--repo-name",
            "tools",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_issue_create_missing_repo_owner_exits_nonzero(monkeypatch) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "token")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "create",
            "new issue",
            "--repo-name",
            "tools",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_issue_create_missing_repo_name_exits_nonzero(monkeypatch) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "token")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "create",
            "new issue",
            "--repo-owner",
            "acme",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_issue_create_invalid_repo_owner_exits_nonzero(monkeypatch) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "token")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "create",
            "new issue",
            "--repo-owner",
            "invalid/owner",
            "--repo-name",
            "tools",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_issue_create_invalid_repo_name_exits_nonzero(monkeypatch) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "token")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "issue",
            "create",
            "new issue",
            "--repo-owner",
            "acme",
            "--repo-name",
            "invalid/repo",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_pr_create_routes_correctly(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}

    async def fake_run_pr_create(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    monkeypatch.setattr(main_module, "run_pr_create", fake_run_pr_create)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "pr",
            "create",
            "--workspace",
            str(tmp_path),
            "--title",
            "My PR Title",
            "--body",
            "PR description here",
            "--base",
            "main",
            "--head",
            "feature-branch",
            "--github-token",
            "ghp_test123",
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["kwargs"]["workspace_dir"] == tmp_path.resolve()
    assert captured["kwargs"]["title"] == "My PR Title"
    assert captured["kwargs"]["body"] == "PR description here"
    assert captured["kwargs"]["base_branch"] == "main"
    assert captured["kwargs"]["head_branch"] == "feature-branch"
    assert captured["kwargs"]["github_token"] == "ghp_test123"


def test_main_pr_create_omitted_title_body_routes_correctly(monkeypatch, tmp_path) -> None:
    """Verify that omitting title/body still routes to run_pr_create (suggest path)."""
    captured: dict[str, Any] = {"ran": False}

    async def fake_run_pr_create(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    monkeypatch.setattr(main_module, "run_pr_create", fake_run_pr_create)
    monkeypatch.setenv("GITHUB_TOKEN", "ghp_test")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "pr",
            "create",
            "--workspace",
            str(tmp_path),
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["kwargs"]["title"] is None
    assert captured["kwargs"]["body"] is None


def test_main_pr_create_missing_subcommand_exits_nonzero(monkeypatch, tmp_path) -> None:
    monkeypatch.setattr(
        sys,
        "argv",
        ["brokk", "pr", "--workspace", str(tmp_path)],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_pr_create_with_explicit_title_body(mock_executor_class, tmp_path) -> None:
    """Verifies run_pr_create skips suggest when title and body are provided."""
    from unittest.mock import AsyncMock

    call_order: list[str] = []
    mock_manager = mock_executor_class.return_value

    async def mock_start():
        call_order.append("start")

    async def mock_create_session(name: str = ""):
        call_order.append("create_session")
        return "session-123"

    async def mock_wait_ready(timeout: float = 30.0):
        call_order.append("wait_ready")
        return True

    async def mock_pr_suggest(**kwargs):
        call_order.append("pr_suggest")
        return {"title": "Suggested", "description": "Suggested desc"}

    async def mock_pr_create(**kwargs):
        call_order.append(f"pr_create:title={kwargs.get('title')}")
        return {"url": "https://github.com/test/repo/pull/42"}

    async def mock_stop():
        call_order.append("stop")

    mock_manager.start = AsyncMock(side_effect=mock_start)
    mock_manager.create_session = AsyncMock(side_effect=mock_create_session)
    mock_manager.wait_ready = AsyncMock(side_effect=mock_wait_ready)
    mock_manager.pr_suggest = AsyncMock(side_effect=mock_pr_suggest)
    mock_manager.pr_create = AsyncMock(side_effect=mock_pr_create)
    mock_manager.stop = AsyncMock(side_effect=mock_stop)

    await main_module.run_pr_create(
        workspace_dir=tmp_path,
        title="Explicit Title",
        body="Explicit Body",
        github_token="ghp_test",
    )

    # pr_suggest should NOT be called when title and body are explicit
    assert "pr_suggest" not in call_order
    assert "pr_create:title=Explicit Title" in call_order
    assert "stop" in call_order


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_pr_create_suggests_when_title_missing(mock_executor_class, tmp_path) -> None:
    """Verifies run_pr_create calls suggest when title is missing."""
    from unittest.mock import AsyncMock

    call_order: list[str] = []
    mock_manager = mock_executor_class.return_value

    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)

    async def mock_pr_suggest(**kwargs):
        call_order.append("pr_suggest")
        return {"title": "Suggested Title", "description": "Suggested Desc"}

    async def mock_pr_create(**kwargs):
        call_order.append(f"pr_create:title={kwargs.get('title')}")
        return {"url": "https://github.com/test/repo/pull/99"}

    mock_manager.pr_suggest = AsyncMock(side_effect=mock_pr_suggest)
    mock_manager.pr_create = AsyncMock(side_effect=mock_pr_create)
    mock_manager.stop = AsyncMock()

    await main_module.run_pr_create(
        workspace_dir=tmp_path,
        title=None,
        body="Explicit Body",
        github_token="ghp_test",
    )

    assert "pr_suggest" in call_order
    # Title should come from suggestion, body was explicit
    assert "pr_create:title=Suggested Title" in call_order


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_pr_create_suggests_when_both_title_and_body_omitted(
    mock_executor_class, tmp_path, capsys
) -> None:
    """Verifies run_pr_create calls suggest with branches and token when both fields omitted."""
    from unittest.mock import AsyncMock

    captured_suggest_args: dict[str, Any] = {}
    captured_create_args: dict[str, Any] = {}
    mock_manager = mock_executor_class.return_value

    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)

    async def mock_pr_suggest(**kwargs):
        captured_suggest_args.update(kwargs)
        return {"title": "Auto-generated Title", "description": "Auto-generated Body"}

    async def mock_pr_create(**kwargs):
        captured_create_args.update(kwargs)
        return {"url": "https://github.com/org/repo/pull/777"}

    mock_manager.pr_suggest = AsyncMock(side_effect=mock_pr_suggest)
    mock_manager.pr_create = AsyncMock(side_effect=mock_pr_create)
    mock_manager.stop = AsyncMock()

    await main_module.run_pr_create(
        workspace_dir=tmp_path,
        title=None,
        body=None,
        base_branch="main",
        head_branch="feature-xyz",
        github_token="ghp_both_omitted_token",
    )

    # Verify pr_suggest was called with resolved branches and token
    mock_manager.pr_suggest.assert_called_once()
    assert captured_suggest_args["source_branch"] == "feature-xyz"
    assert captured_suggest_args["target_branch"] == "main"
    assert captured_suggest_args["github_token"] == "ghp_both_omitted_token"

    # Verify pr_create received the suggested title and description
    mock_manager.pr_create.assert_called_once()
    assert captured_create_args["title"] == "Auto-generated Title"
    assert captured_create_args["body"] == "Auto-generated Body"
    assert captured_create_args["source_branch"] == "feature-xyz"
    assert captured_create_args["target_branch"] == "main"
    assert captured_create_args["github_token"] == "ghp_both_omitted_token"

    # Verify PR URL is printed
    captured = capsys.readouterr()
    assert "https://github.com/org/repo/pull/777" in captured.out

    mock_manager.stop.assert_awaited_once()


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_pr_create_executor_error_exits_nonzero(
    mock_executor_class, tmp_path, capsys
) -> None:
    """Verifies run_pr_create exits non-zero on executor error."""
    from unittest.mock import AsyncMock

    from brokk_code.executor import ExecutorError

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.pr_create = AsyncMock(side_effect=ExecutorError("GitHub API error"))
    mock_manager.stop = AsyncMock()

    with pytest.raises(SystemExit) as exc:
        await main_module.run_pr_create(
            workspace_dir=tmp_path,
            title="Test",
            body="Test body",
            github_token="ghp_test",
        )

    assert exc.value.code == 1
    captured = capsys.readouterr()
    assert "Executor error" in captured.err
    assert "GitHub API error" in captured.err
    mock_manager.stop.assert_awaited_once()


def test_main_commit_routes_correctly(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}

    async def fake_run_commit(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    monkeypatch.setattr(main_module, "run_commit", fake_run_commit)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "commit",
            "Fix the bug",
            "--workspace",
            str(tmp_path),
            "--vendor",
            "Anthropic",
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["kwargs"]["workspace_dir"] == tmp_path.resolve()
    assert captured["kwargs"]["message"] == "Fix the bug"
    assert captured["kwargs"]["vendor"] == "Anthropic"


def test_main_commit_no_message_routes_correctly(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}

    async def fake_run_commit(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    monkeypatch.setattr(main_module, "run_commit", fake_run_commit)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "commit",
            "--workspace",
            str(tmp_path),
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["kwargs"]["workspace_dir"] == tmp_path.resolve()
    assert captured["kwargs"]["message"] is None


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_commit_calls_lifecycle_in_order(mock_executor_class, tmp_path) -> None:
    """Verifies that run_commit follows the correct lifecycle order."""
    from unittest.mock import AsyncMock

    call_order: list[str] = []
    mock_manager = mock_executor_class.return_value

    async def mock_start():
        call_order.append("start")

    async def mock_create_session(name: str = ""):
        call_order.append("create_session")
        return "session-123"

    async def mock_wait_ready(timeout: float = 30.0):
        call_order.append("wait_ready")
        return True

    async def mock_commit_context(message=None):
        call_order.append(f"commit_context:{message}")
        return {"commitId": "abc123", "firstLine": "Test commit"}

    async def mock_stop():
        call_order.append("stop")

    mock_manager.start = AsyncMock(side_effect=mock_start)
    mock_manager.create_session = AsyncMock(side_effect=mock_create_session)
    mock_manager.wait_ready = AsyncMock(side_effect=mock_wait_ready)
    mock_manager.commit_context = AsyncMock(side_effect=mock_commit_context)
    mock_manager.stop = AsyncMock(side_effect=mock_stop)

    await main_module.run_commit(
        workspace_dir=tmp_path,
        message="My commit message",
    )

    assert "start" in call_order
    assert "create_session" in call_order
    assert "wait_ready" in call_order
    assert "commit_context:My commit message" in call_order
    assert "stop" in call_order

    start_idx = call_order.index("start")
    create_session_idx = call_order.index("create_session")
    wait_ready_idx = call_order.index("wait_ready")
    commit_idx = call_order.index("commit_context:My commit message")
    stop_idx = call_order.index("stop")

    assert start_idx < create_session_idx
    assert create_session_idx < wait_ready_idx
    assert wait_ready_idx < commit_idx
    assert commit_idx < stop_idx


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_commit_no_changes(mock_executor_class, tmp_path, capsys) -> None:
    """Verifies that run_commit handles no changes case."""
    from unittest.mock import AsyncMock

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.commit_context = AsyncMock(return_value={"status": "no_changes"})
    mock_manager.stop = AsyncMock()

    await main_module.run_commit(workspace_dir=tmp_path, message=None)

    captured = capsys.readouterr()
    assert "No uncommitted changes" in captured.out
    mock_manager.stop.assert_awaited_once()


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_commit_success_output(mock_executor_class, tmp_path, capsys) -> None:
    """Verifies that run_commit prints commit info on success."""
    from unittest.mock import AsyncMock

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.commit_context = AsyncMock(
        return_value={"commitId": "abc1234567890", "firstLine": "Fix parser bug"}
    )
    mock_manager.stop = AsyncMock()

    await main_module.run_commit(workspace_dir=tmp_path, message="Fix parser bug")

    captured = capsys.readouterr()
    assert "abc1234" in captured.out
    assert "Fix parser bug" in captured.out
    mock_manager.stop.assert_awaited_once()


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_commit_executor_error_exits_nonzero(
    mock_executor_class, tmp_path, capsys
) -> None:
    """Verifies that run_commit exits non-zero on executor error."""
    from unittest.mock import AsyncMock

    from brokk_code.executor import ExecutorError

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.commit_context = AsyncMock(side_effect=ExecutorError("Git error"))
    mock_manager.stop = AsyncMock()

    with pytest.raises(SystemExit) as exc:
        await main_module.run_commit(workspace_dir=tmp_path, message="test")

    assert exc.value.code == 1
    captured = capsys.readouterr()
    assert "Executor error" in captured.err
    assert "Git error" in captured.err
    mock_manager.stop.assert_awaited_once()


def test_main_pr_review_routes_correctly(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}

    async def fake_run_pr_review_job(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    monkeypatch.setattr(main_module, "run_pr_review_job", fake_run_pr_review_job)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "pr",
            "review",
            "--pr-number",
            "42",
            "--workspace",
            str(tmp_path),
            "--github-token",
            "ghp_test",
            "--repo-owner",
            "acme",
            "--repo-name",
            "tools",
            "--planner-model",
            "custom-model",
            "--verbose",
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["kwargs"]["workspace_dir"] == tmp_path.resolve()
    assert captured["kwargs"]["pr_number"] == 42
    assert captured["kwargs"]["github_token"] == "ghp_test"
    assert captured["kwargs"]["repo_owner"] == "acme"
    assert captured["kwargs"]["repo_name"] == "tools"
    assert captured["kwargs"]["planner_model"] == "custom-model"
    assert captured["kwargs"]["verbose"] is True


def test_main_pr_review_missing_pr_number_exits_nonzero(monkeypatch, tmp_path) -> None:
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "pr",
            "review",
            "--github-token",
            "ghp_test",
            "--repo-owner",
            "acme",
            "--repo-name",
            "tools",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code != 0


def test_main_pr_review_missing_github_token_exits_nonzero(monkeypatch, tmp_path) -> None:
    monkeypatch.delenv("GITHUB_TOKEN", raising=False)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "pr",
            "review",
            "--pr-number",
            "42",
            "--workspace",
            str(tmp_path),
            "--repo-owner",
            "acme",
            "--repo-name",
            "tools",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 1


def test_main_pr_review_missing_repo_owner_without_inference_exits_nonzero(
    monkeypatch, tmp_path
) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "token")
    monkeypatch.setattr(main_module, "infer_github_repo_from_remote", lambda _: (None, None))
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "pr",
            "review",
            "--pr-number",
            "42",
            "--workspace",
            str(tmp_path),
            "--repo-name",
            "tools",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 1


def test_main_pr_review_missing_repo_name_without_inference_exits_nonzero(
    monkeypatch, tmp_path
) -> None:
    monkeypatch.setenv("GITHUB_TOKEN", "token")
    monkeypatch.setattr(main_module, "infer_github_repo_from_remote", lambda _: (None, None))
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "pr",
            "review",
            "--pr-number",
            "42",
            "--workspace",
            str(tmp_path),
            "--repo-owner",
            "acme",
        ],
    )

    with pytest.raises(SystemExit) as exc:
        main_module.main()

    assert exc.value.code == 1


def test_main_pr_review_infers_repo_from_https_remote(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}

    async def fake_run_pr_review_job(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    def fake_subprocess_run(cmd, **kwargs):
        class FakeResult:
            returncode = 0
            stdout = "https://github.com/inferred-owner/inferred-repo.git\n"
            stderr = ""

        return FakeResult()

    monkeypatch.setattr(main_module, "run_pr_review_job", fake_run_pr_review_job)
    monkeypatch.setattr(subprocess, "run", fake_subprocess_run)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "pr",
            "review",
            "--pr-number",
            "42",
            "--workspace",
            str(tmp_path),
            "--github-token",
            "ghp_test",
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["kwargs"]["repo_owner"] == "inferred-owner"
    assert captured["kwargs"]["repo_name"] == "inferred-repo"


def test_main_pr_review_infers_repo_from_ssh_remote(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}

    async def fake_run_pr_review_job(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    def fake_subprocess_run(cmd, **kwargs):
        class FakeResult:
            returncode = 0
            stdout = "git@github.com:ssh-owner/ssh-repo.git\n"
            stderr = ""

        return FakeResult()

    monkeypatch.setattr(main_module, "run_pr_review_job", fake_run_pr_review_job)
    monkeypatch.setattr(subprocess, "run", fake_subprocess_run)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "pr",
            "review",
            "--pr-number",
            "42",
            "--workspace",
            str(tmp_path),
            "--github-token",
            "ghp_test",
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["kwargs"]["repo_owner"] == "ssh-owner"
    assert captured["kwargs"]["repo_name"] == "ssh-repo"


def test_main_pr_review_explicit_params_override_inference(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}

    async def fake_run_pr_review_job(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    def fake_subprocess_run(cmd, **kwargs):
        class FakeResult:
            returncode = 0
            stdout = "https://github.com/inferred-owner/inferred-repo.git\n"
            stderr = ""

        return FakeResult()

    monkeypatch.setattr(main_module, "run_pr_review_job", fake_run_pr_review_job)
    monkeypatch.setattr(subprocess, "run", fake_subprocess_run)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "pr",
            "review",
            "--pr-number",
            "42",
            "--workspace",
            str(tmp_path),
            "--github-token",
            "ghp_test",
            "--repo-owner",
            "explicit-owner",
            "--repo-name",
            "explicit-repo",
        ],
    )

    main_module.main()

    assert captured["ran"] is True
    assert captured["kwargs"]["repo_owner"] == "explicit-owner"
    assert captured["kwargs"]["repo_name"] == "explicit-repo"


def test_main_pr_review_respects_env_github_token(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}

    async def fake_run_pr_review_job(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    monkeypatch.setattr(main_module, "run_pr_review_job", fake_run_pr_review_job)
    monkeypatch.setenv("GITHUB_TOKEN", "env-token")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "pr",
            "review",
            "--pr-number",
            "42",
            "--workspace",
            str(tmp_path),
            "--repo-owner",
            "acme",
            "--repo-name",
            "tools",
        ],
    )

    main_module.main()

    assert captured["kwargs"]["github_token"] == "env-token"


def test_main_pr_review_uses_default_planner_model(monkeypatch, tmp_path) -> None:
    captured: dict[str, Any] = {"ran": False}

    async def fake_run_pr_review_job(**kwargs: Any) -> None:
        captured["kwargs"] = kwargs
        captured["ran"] = True

    monkeypatch.setattr(main_module, "run_pr_review_job", fake_run_pr_review_job)
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "brokk",
            "pr",
            "review",
            "--pr-number",
            "42",
            "--workspace",
            str(tmp_path),
            "--github-token",
            "ghp_test",
            "--repo-owner",
            "acme",
            "--repo-name",
            "tools",
        ],
    )

    main_module.main()

    assert captured["kwargs"]["planner_model"] == "gpt-5.1"


def test_infer_github_repo_from_remote_https_format(monkeypatch, tmp_path) -> None:
    def fake_subprocess_run(cmd, **kwargs):
        class FakeResult:
            returncode = 0
            stdout = "https://github.com/test-owner/test-repo.git\n"
            stderr = ""

        return FakeResult()

    monkeypatch.setattr(subprocess, "run", fake_subprocess_run)

    owner, repo = git_utils_module.infer_github_repo_from_remote(tmp_path)

    assert owner == "test-owner"
    assert repo == "test-repo"


def test_infer_github_repo_from_remote_https_no_git_suffix(monkeypatch, tmp_path) -> None:
    def fake_subprocess_run(cmd, **kwargs):
        class FakeResult:
            returncode = 0
            stdout = "https://github.com/test-owner/test-repo\n"
            stderr = ""

        return FakeResult()

    monkeypatch.setattr(subprocess, "run", fake_subprocess_run)

    owner, repo = git_utils_module.infer_github_repo_from_remote(tmp_path)

    assert owner == "test-owner"
    assert repo == "test-repo"


def test_infer_github_repo_from_remote_ssh_format(monkeypatch, tmp_path) -> None:
    def fake_subprocess_run(cmd, **kwargs):
        class FakeResult:
            returncode = 0
            stdout = "git@github.com:ssh-owner/ssh-repo.git\n"
            stderr = ""

        return FakeResult()

    monkeypatch.setattr(subprocess, "run", fake_subprocess_run)

    owner, repo = git_utils_module.infer_github_repo_from_remote(tmp_path)

    assert owner == "ssh-owner"
    assert repo == "ssh-repo"


def test_infer_github_repo_from_remote_ssh_no_git_suffix(monkeypatch, tmp_path) -> None:
    def fake_subprocess_run(cmd, **kwargs):
        class FakeResult:
            returncode = 0
            stdout = "git@github.com:ssh-owner/ssh-repo\n"
            stderr = ""

        return FakeResult()

    monkeypatch.setattr(subprocess, "run", fake_subprocess_run)

    owner, repo = git_utils_module.infer_github_repo_from_remote(tmp_path)

    assert owner == "ssh-owner"
    assert repo == "ssh-repo"


def test_infer_github_repo_from_remote_non_github_returns_none(monkeypatch, tmp_path) -> None:
    def fake_subprocess_run(cmd, **kwargs):
        class FakeResult:
            returncode = 0
            stdout = "https://gitlab.com/owner/repo.git\n"
            stderr = ""

        return FakeResult()

    monkeypatch.setattr(subprocess, "run", fake_subprocess_run)

    owner, repo = git_utils_module.infer_github_repo_from_remote(tmp_path)

    assert owner is None
    assert repo is None


def test_infer_github_repo_from_remote_git_failure_returns_none(monkeypatch, tmp_path) -> None:
    def fake_subprocess_run(cmd, **kwargs):
        class FakeResult:
            returncode = 1
            stdout = ""
            stderr = "fatal: not a git repository"

        return FakeResult()

    monkeypatch.setattr(subprocess, "run", fake_subprocess_run)

    owner, repo = git_utils_module.infer_github_repo_from_remote(tmp_path)

    assert owner is None
    assert repo is None


def test_infer_github_repo_from_remote_empty_stdout_returns_none(monkeypatch, tmp_path) -> None:
    def fake_subprocess_run(cmd, **kwargs):
        class FakeResult:
            returncode = 0
            stdout = ""
            stderr = ""

        return FakeResult()

    monkeypatch.setattr(subprocess, "run", fake_subprocess_run)

    owner, repo = git_utils_module.infer_github_repo_from_remote(tmp_path)

    assert owner is None
    assert repo is None


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_pr_review_job_calls_submit_pr_review_job(mock_executor_class, tmp_path) -> None:
    from unittest.mock import AsyncMock

    call_order: list[str] = []
    mock_manager = mock_executor_class.return_value

    async def mock_start():
        call_order.append("start")

    async def mock_create_session(name: str = ""):
        call_order.append(f"create_session:{name}")
        return "session-123"

    async def mock_wait_ready(timeout: float = 30.0):
        call_order.append("wait_ready")
        return True

    async def mock_submit_pr_review_job(**kwargs):
        call_order.append(f"submit_pr_review_job:{kwargs}")
        return "pr-review-job-456"

    async def mock_stream_events(job_id: str):
        yield {"type": "STATE_CHANGE", "data": {"state": "COMPLETED"}}

    async def mock_stop():
        call_order.append("stop")

    mock_manager.start = AsyncMock(side_effect=mock_start)
    mock_manager.create_session = AsyncMock(side_effect=mock_create_session)
    mock_manager.wait_ready = AsyncMock(side_effect=mock_wait_ready)
    mock_manager.submit_pr_review_job = AsyncMock(side_effect=mock_submit_pr_review_job)
    mock_manager.stream_events = mock_stream_events
    mock_manager.stop = AsyncMock(side_effect=mock_stop)

    await main_module.run_pr_review_job(
        workspace_dir=tmp_path,
        pr_number=42,
        github_token="ghp_test",
        repo_owner="test-owner",
        repo_name="test-repo",
        planner_model="gpt-4",
    )

    assert "start" in call_order
    assert any("create_session:PR Review #42" in c for c in call_order)
    assert "wait_ready" in call_order
    assert any("submit_pr_review_job" in c for c in call_order)

    submit_call = [c for c in call_order if "submit_pr_review_job" in c][0]
    assert "planner_model" in submit_call
    assert "gpt-4" in submit_call
    assert "ghp_test" in submit_call
    assert "test-owner" in submit_call
    assert "test-repo" in submit_call
    assert "42" in submit_call


@pytest.mark.asyncio
@patch("brokk_code.executor.ExecutorManager")
async def test_run_pr_review_job_exits_nonzero_on_failed_state(
    mock_executor_class, tmp_path, capsys
) -> None:
    from unittest.mock import AsyncMock

    mock_manager = mock_executor_class.return_value
    mock_manager.start = AsyncMock()
    mock_manager.create_session = AsyncMock(return_value="session-123")
    mock_manager.wait_ready = AsyncMock(return_value=True)
    mock_manager.submit_pr_review_job = AsyncMock(return_value="job-456")

    async def mock_stream_events(job_id: str):
        yield {"type": "ERROR", "data": {"message": "GitHub API error"}}
        yield {"type": "STATE_CHANGE", "data": {"state": "FAILED"}}

    mock_manager.stream_events = mock_stream_events
    mock_manager.stop = AsyncMock()

    with pytest.raises(SystemExit) as exc:
        await main_module.run_pr_review_job(
            workspace_dir=tmp_path,
            pr_number=42,
            github_token="ghp_test",
            repo_owner="test-owner",
            repo_name="test-repo",
            planner_model="gpt-4",
        )

    assert exc.value.code == 1
    captured = capsys.readouterr()
    assert "PR review job ended with state FAILED" in captured.err


def test_install_calls_ensure_jbang_ready(monkeypatch, tmp_path) -> None:
    """install command calls ensure_jbang_ready() directly (not _resolve_jbang_for_install)."""
    ensure_called = {"n": 0}

    def fake_ensure_jbang_ready() -> str:
        ensure_called["n"] += 1
        return "/usr/bin/jbang"

    def fake_configure_zed_acp_settings(*, force=False, uvx_command=None):
        return tmp_path / "zed.json"

    monkeypatch.setattr(main_module, "ensure_jbang_ready", fake_ensure_jbang_ready)
    monkeypatch.setattr(main_module, "_run_install_prefetch", lambda _commands: None)
    monkeypatch.setattr(main_module, "configure_zed_acp_settings", fake_configure_zed_acp_settings)
    monkeypatch.setattr(
        main_module,
        "_build_install_prefetch_commands",
        lambda **kwargs: [],
    )
    monkeypatch.setattr(main_module, "sys", __import__("sys"))
    monkeypatch.setattr(
        __import__("sys"),
        "argv",
        ["brokk", "install", "zed"],
    )

    main_module.main()

    assert ensure_called["n"] == 1



from typing import Dict, List

from brokk_code import __version__


def get_braille_icon() -> str:
    """
    Returns a multi-line string representing the Brokk icon using Braille characters.
    Uses characters in the U+2800–U+28FF range.
    Depicts the Viking helmet logo: two curved horns, dome, central bar, and banded lower section.
    """
    return (
        "⠀⠀⣷⡀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢠⡾⠀⠀\n"
        "⠀⢀⣿⡇⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢸⣷⡀⠀\n"
        "⢠⣸⡟⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⠀⢻⣇⡄\n"
        "⣿⣷⡀⠀⠀⠀⠀⡀⣤⣤⢀⠀⠀⠀⠀⢀⣾⣿\n"
        "⢿⣿⣿⣷⡤⣠⣾⡇⣿⣿⢸⣷⣄⢤⣾⣿⣿⡿\n"
        "⠀⠻⢿⡟⣴⣿⣿⡇⣿⣿⢸⣿⣿⣮⢻⡿⠟⠀\n"
        "⠀⠀⠀⠐⢿⣿⣿⡇⣿⣿⢸⣿⣿⡿⠂⠀⠀⠀\n"
        "⠀⠀⠀⣝⠷⣬⣭⡁⣿⣿⢨⣭⣵⠾⣫⠀⠀⠀\n"
        "⠀⠀⠀⠈⠛⢾⣭⣥⣿⣿⣬⣭⡶⠛⠁⠀⠀⠀\n"
        "⠀⠀⠀⠀⠀⠀⠀⢸⣿⣿⡇⠀⠀⠀⠀⠀⠀⠀\n"
        "⠀⠀⠀⠀⠀⠀⠀⠘⢝⡫⠃⠀⠀⠀⠀⠀⠀⠀"
    )


def build_welcome_message(
    commands: List[Dict[str, str]], latest_pypi_version: str | None = None
) -> str:
    """
    Constructs the branded welcome/onboarding message as a Markdown string.

    Args:
        commands: The list of supported slash commands from BrokkApp.get_slash_commands().
        latest_pypi_version: The latest version available on PyPI, if known.

    Returns:
        A Markdown-formatted string for display in the ChatPanel.
    """
    version_info = f"v{__version__}"
    if latest_pypi_version and latest_pypi_version != __version__:
        version_info += f" (Latest: {latest_pypi_version})"

    description = (
        f"# Welcome to Brokk {version_info}\n\n"
        "Brokk is a code intelligence agent designed for high-precision **context engineering**."
    )

    context_eng = (
        "### Context Engineering\n"
        "In Brokk, your workspace is managed via a living context. "
        "AI performance depends on exactly what code is visible to the model. "
        "You can prune, pin, or focus specific files and methods to optimize results."
    )

    workflows = (
        "### Key Workflows\n"
        "- **`/context`**: Toggle the Context Panel to manage what the AI sees. "
        "Use it to drop unnecessary fragments and keep your token usage efficient.\n"
        "- **`/task`**: Open the Task List to track your current objectives. "
        "The AI uses this list to stay aligned with your goals.\n"
        "- **`@mentions`**: Type `@` followed by a file or class name in the chat to "
        "immediately attach that entity to your context."
    )

    signup = "\n\nNeed a token? Go to https://brokk.ai/ to sign up and get one."

    return f"{description}\n\n{context_eng}\n\n{workflows}{signup}"



"""Tests for the /pr slash command."""

from pathlib import Path
from unittest.mock import AsyncMock, MagicMock

import pytest

from brokk_code.app import BrokkApp, PrCreateModalScreen


def test_pr_in_slash_command_catalog():
    """Verify /pr is listed in the slash commands catalog."""
    commands = BrokkApp.get_slash_commands()
    command_names = [c["command"] for c in commands]
    assert "/pr" in command_names

    pr_cmd = next(c for c in commands if c["command"] == "/pr")
    assert "description" in pr_cmd
    assert pr_cmd["description"]


def test_handle_command_pr_requires_executor_ready(tmp_path: Path):
    """Verify /pr shows error when executor is not ready."""
    mock_executor = MagicMock()
    mock_executor.workspace_dir = tmp_path

    app = BrokkApp(workspace_dir=tmp_path, executor=mock_executor)
    app._executor_ready = False

    # Mock the chat panel
    mock_chat = MagicMock()
    app.query_one = MagicMock(return_value=mock_chat)

    app._handle_command("/pr")

    mock_chat.add_system_message.assert_called_once()
    call_args = mock_chat.add_system_message.call_args
    assert "not ready" in call_args[0][0].lower()
    assert call_args[1].get("level") == "ERROR"


@pytest.mark.asyncio
async def test_create_pull_request_fetches_sessions_and_suggestion(tmp_path: Path):
    """Verify _create_pull_request fetches overlapping sessions and PR suggestion from executor."""
    mock_executor = MagicMock()
    mock_executor.workspace_dir = tmp_path
    mock_executor.pr_sessions = AsyncMock(
        return_value={
            "sessions": [
                {"id": "session-1", "name": "Session One", "taskCount": 3},
                {"id": "session-2", "name": "Session Two", "taskCount": 5},
            ],
            "sourceBranch": "feature-branch",
            "targetBranch": "main",
        }
    )
    mock_executor.pr_suggest = AsyncMock(
        return_value={
            "title": "Add new feature",
            "description": "This PR adds a cool feature",
            "sourceBranch": "feature-branch",
            "targetBranch": "main",
        }
    )

    app = BrokkApp(workspace_dir=tmp_path, executor=mock_executor)
    app._executor_ready = True
    app.current_branch = "feature-branch"

    mock_chat = MagicMock()
    mock_chat.add_system_message = MagicMock()
    mock_chat.set_job_running = MagicMock()
    app._maybe_chat = MagicMock(return_value=mock_chat)

    # Mock push_screen to capture the modal
    pushed_screen = {}

    def capture_push_screen(screen, callback=None):
        pushed_screen["screen"] = screen
        pushed_screen["callback"] = callback

    app.push_screen = MagicMock(side_effect=capture_push_screen)

    await app._create_pull_request(base_branch="main")

    # Verify overlapping sessions were fetched with correct branches
    mock_executor.pr_sessions.assert_called_once_with(
        source_branch="feature-branch",
        target_branch="main",
    )

    # Verify pr_suggest was called with ALL overlapping session IDs (Swing behavior)
    mock_executor.pr_suggest.assert_called_once_with(
        source_branch="feature-branch",
        target_branch="main",
        session_ids=["session-1", "session-2"],
    )

    # Verify modal was pushed with correct data including sessions
    assert "screen" in pushed_screen
    screen = pushed_screen["screen"]
    assert isinstance(screen, PrCreateModalScreen)
    assert screen._suggested_title == "Add new feature"
    assert screen._suggested_body == "This PR adds a cool feature"
    assert screen._source_branch == "feature-branch"
    assert screen._target_branch == "main"
    assert len(screen._sessions) == 2
    # All overlapping sessions should be pre-selected
    assert "session-1" in screen._selected_session_ids
    assert "session-2" in screen._selected_session_ids


@pytest.mark.asyncio
async def test_do_create_pr_success(tmp_path: Path):
    """Verify _do_create_pr handles successful PR creation."""
    mock_executor = MagicMock()
    mock_executor.workspace_dir = tmp_path
    mock_executor.pr_create = AsyncMock(
        return_value={"url": "https://github.com/owner/repo/pull/42"}
    )

    app = BrokkApp(workspace_dir=tmp_path, executor=mock_executor)
    app._executor_ready = True

    mock_chat = MagicMock()
    mock_chat.add_system_message = MagicMock()
    mock_chat.add_system_message_markup = MagicMock()
    mock_chat.set_job_running = MagicMock()
    app._maybe_chat = MagicMock(return_value=mock_chat)
    app._refresh_context_panel = AsyncMock()

    await app._do_create_pr(
        title="My PR",
        body="PR description",
        source_branch="feature",
        target_branch="main",
    )

    mock_executor.pr_create.assert_called_once_with(
        title="My PR",
        body="PR description",
        source_branch="feature",
        target_branch="main",
        session_ids=None,
    )

    # Check that the PR URL was shown
    mock_chat.add_system_message_markup.assert_called_once()
    call_arg = mock_chat.add_system_message_markup.call_args[0][0]
    assert "https://github.com/owner/repo/pull/42" in call_arg


@pytest.mark.asyncio
async def test_do_create_pr_with_session_ids(tmp_path: Path):
    """Verify _do_create_pr forwards session_ids to executor."""
    mock_executor = MagicMock()
    mock_executor.workspace_dir = tmp_path
    mock_executor.pr_create = AsyncMock(
        return_value={"url": "https://github.com/owner/repo/pull/42"}
    )

    app = BrokkApp(workspace_dir=tmp_path, executor=mock_executor)
    app._executor_ready = True

    mock_chat = MagicMock()
    mock_chat.add_system_message = MagicMock()
    mock_chat.add_system_message_markup = MagicMock()
    mock_chat.set_job_running = MagicMock()
    app._maybe_chat = MagicMock(return_value=mock_chat)
    app._refresh_context_panel = AsyncMock()

    await app._do_create_pr(
        title="My PR",
        body="PR description",
        source_branch="feature",
        target_branch="main",
        session_ids=["session-1", "session-2"],
    )

    mock_executor.pr_create.assert_called_once_with(
        title="My PR",
        body="PR description",
        source_branch="feature",
        target_branch="main",
        session_ids=["session-1", "session-2"],
    )


@pytest.mark.asyncio
async def test_do_create_pr_error(tmp_path: Path):
    """Verify _do_create_pr handles errors gracefully."""
    mock_executor = MagicMock()
    mock_executor.workspace_dir = tmp_path
    mock_executor.pr_create = AsyncMock(side_effect=Exception("GitHub API error"))

    app = BrokkApp(workspace_dir=tmp_path, executor=mock_executor)
    app._executor_ready = True

    mock_chat = MagicMock()
    mock_chat.add_system_message = MagicMock()
    mock_chat.set_job_running = MagicMock()
    app._maybe_chat = MagicMock(return_value=mock_chat)

    await app._do_create_pr(
        title="My PR",
        body="PR description",
        source_branch="feature",
        target_branch="main",
    )

    # Find the error message call
    error_calls = [
        call
        for call in mock_chat.add_system_message.call_args_list
        if call[1].get("level") == "ERROR"
    ]
    assert len(error_calls) == 1
    assert "failed" in error_calls[0][0][0].lower()


@pytest.mark.asyncio
async def test_create_pull_request_suggestion_error(tmp_path: Path):
    """Verify _create_pull_request handles suggestion errors gracefully."""
    mock_executor = MagicMock()
    mock_executor.workspace_dir = tmp_path
    mock_executor.pr_sessions = AsyncMock(
        return_value={"sessions": [], "sourceBranch": "feature-branch", "targetBranch": "main"}
    )
    mock_executor.pr_suggest = AsyncMock(side_effect=Exception("Network error"))

    app = BrokkApp(workspace_dir=tmp_path, executor=mock_executor)
    app._executor_ready = True
    app.current_branch = "feature-branch"

    mock_chat = MagicMock()
    mock_chat.add_system_message = MagicMock()
    mock_chat.set_job_running = MagicMock()
    app._maybe_chat = MagicMock(return_value=mock_chat)

    await app._create_pull_request()

    # Find the error message call
    error_calls = [
        call
        for call in mock_chat.add_system_message.call_args_list
        if call[1].get("level") == "ERROR"
    ]
    assert len(error_calls) == 1
    assert "suggestion" in error_calls[0][0][0].lower()
    # Ensure "sessions" is NOT in the error message (distinguishing from sessions failure)
    assert "sessions" not in error_calls[0][0][0].lower()

    # Verify set_job_running(False) was called after the error
    mock_chat.set_job_running.assert_called()
    assert mock_chat.set_job_running.call_args_list[-1] == (((False,), {}))


@pytest.mark.asyncio
async def test_create_pull_request_sessions_error(tmp_path: Path):
    """Verify _create_pull_request handles pr_sessions errors gracefully with specific message."""
    mock_executor = MagicMock()
    mock_executor.workspace_dir = tmp_path
    mock_executor.pr_sessions = AsyncMock(side_effect=Exception("Network error"))
    mock_executor.pr_suggest = AsyncMock()

    app = BrokkApp(workspace_dir=tmp_path, executor=mock_executor)
    app._executor_ready = True
    app.current_branch = "feature-branch"

    mock_chat = MagicMock()
    mock_chat.add_system_message = MagicMock()
    mock_chat.set_job_running = MagicMock()
    app._maybe_chat = MagicMock(return_value=mock_chat)

    await app._create_pull_request()

    # Find the error message call
    error_calls = [
        call
        for call in mock_chat.add_system_message.call_args_list
        if call[1].get("level") == "ERROR"
    ]
    assert len(error_calls) == 1
    # Error message should mention "sessions", not "suggestion"
    assert "sessions" in error_calls[0][0][0].lower()
    assert "suggestion" not in error_calls[0][0][0].lower()

    # pr_suggest should NOT have been called since pr_sessions failed
    mock_executor.pr_suggest.assert_not_called()

    # Verify set_job_running(False) was called after the error
    mock_chat.set_job_running.assert_called()
    assert mock_chat.set_job_running.call_args_list[-1] == (((False,), {}))


def test_pr_create_modal_screen_with_sessions():
    """Verify PrCreateModalScreen can be instantiated with session parameters."""
    sessions = [
        {"id": "sess-1", "name": "Session One"},
        {"id": "sess-2", "name": "Session Two"},
    ]
    screen = PrCreateModalScreen(
        suggested_title="Test Title",
        suggested_body="Test Body",
        source_branch="feature",
        target_branch="main",
        sessions=sessions,
        selected_session_ids=["sess-1"],
    )
    assert screen._suggested_title == "Test Title"
    assert screen._suggested_body == "Test Body"
    assert screen._source_branch == "feature"
    assert screen._target_branch == "main"
    assert len(screen._sessions) == 2
    assert "sess-1" in screen._selected_session_ids
    assert "sess-2" not in screen._selected_session_ids
    # Verify session ID order is preserved
    assert screen._session_id_order == ["sess-1", "sess-2"]



import time
from pathlib import Path

from brokk_code.runtime_utils import find_dev_jar


def test_find_dev_jar_finds_runnable_jar(tmp_path: Path):
    repo_root = tmp_path / "repo"
    libs_dir = repo_root / "app" / "build" / "libs"
    libs_dir.mkdir(parents=True)
    (repo_root / "gradlew").write_text("")

    runnable_jar = libs_dir / "brokk-1.0.0.jar"
    runnable_jar.write_text("dummy")

    assert find_dev_jar(repo_root) == runnable_jar


def test_find_dev_jar_returns_none_when_only_sources_and_javadoc_exist(tmp_path: Path):
    repo_root = tmp_path / "repo"
    libs_dir = repo_root / "app" / "build" / "libs"
    libs_dir.mkdir(parents=True)
    (repo_root / "gradlew").write_text("")

    (libs_dir / "brokk-1.0.0-sources.jar").write_text("sources")
    (libs_dir / "brokk-1.0.0-javadoc.jar").write_text("javadoc")

    assert find_dev_jar(repo_root) is None


def test_find_dev_jar_ignores_sources_and_javadoc_when_runnable_jar_exists(tmp_path: Path):
    repo_root = tmp_path / "repo"
    libs_dir = repo_root / "app" / "build" / "libs"
    libs_dir.mkdir(parents=True)
    (repo_root / "gradlew").write_text("")

    (libs_dir / "brokk-1.0.0-sources.jar").write_text("sources")
    (libs_dir / "brokk-1.0.0-javadoc.jar").write_text("javadoc")

    time.sleep(0.01)
    runnable_jar = libs_dir / "brokk-1.0.0.jar"
    runnable_jar.write_text("runnable")

    assert find_dev_jar(repo_root) == runnable_jar


def test_find_dev_jar_allows_shadow_and_all_jars(tmp_path: Path):
    repo_root = tmp_path / "repo"
    libs_dir = repo_root / "app" / "build" / "libs"
    libs_dir.mkdir(parents=True)
    (repo_root / "gradlew").write_text("")

    shadow_jar = libs_dir / "brokk-1.0.0-shadow.jar"
    shadow_jar.write_text("shadow")
    assert find_dev_jar(repo_root) == shadow_jar

    time.sleep(0.01)
    all_jar = libs_dir / "brokk-1.0.0-all.jar"
    all_jar.write_text("all")
    assert find_dev_jar(repo_root) == all_jar


def test_find_dev_jar_prefers_newest_mtime(tmp_path: Path):
    repo_root = tmp_path / "repo"
    libs_dir = repo_root / "app" / "build" / "libs"
    libs_dir.mkdir(parents=True)
    (repo_root / "gradlew").write_text("")

    old_jar = libs_dir / "brokk-1.0.0.jar"
    old_jar.write_text("old")

    time.sleep(0.01)

    new_jar = libs_dir / "brokk-1.1.0.jar"
    new_jar.write_text("new")

    assert find_dev_jar(repo_root) == new_jar


def test_find_dev_jar_returns_none_when_no_jar(tmp_path: Path):
    repo_root = tmp_path / "repo"
    repo_root.mkdir()
    (repo_root / "gradlew").write_text("")
    # No libs dir
    assert find_dev_jar(repo_root) is None


def test_find_dev_jar_searches_upward(tmp_path: Path):
    repo_root = tmp_path / "repo"
    nested = repo_root / "a" / "b" / "c"
    nested.mkdir(parents=True)

    libs_dir = repo_root / "app" / "build" / "libs"
    libs_dir.mkdir(parents=True)
    (repo_root / "gradlew").write_text("")

    runnable_jar = libs_dir / "brokk-1.0.0.jar"
    runnable_jar.write_text("dummy")

    assert find_dev_jar(nested) == runnable_jar



import pathlib
from pathlib import Path
from typing import Optional

from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widgets import Static

from brokk_code.token_format import format_token_count


class StatusLine(Horizontal):
    """A compact status bar containing metadata and job progress.

    Displays:
      - mode, model, reasoning, workspace (via MetadataLabel)
      - spinner and elapsed timer (via JobProgress)
    """

    SEPARATOR = " • "

    DEFAULT_CSS = """
    StatusLine {
        height: 1;
        padding: 0 1;
        color: $text-disabled;
        layout: horizontal;
    }
    #status-metadata {
        width: 1fr;
        min-width: 10;
    }
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._mode = "unknown"
        self._model = "unknown"
        self._reasoning = "unknown"
        self._workspace = "unknown"
        self._branch = "unknown"
        self._turn_cost: Optional[float] = None
        self._session_cost: Optional[float] = None
        self._fragment_description: Optional[str] = None
        self._fragment_size: Optional[int] = None
        self._metadata: Optional[Static] = None

    def compose(self) -> ComposeResult:
        yield Static(id="status-metadata")

    def on_mount(self) -> None:
        try:
            self._metadata = self.query_one("#status-metadata", Static)
        except Exception:
            self._metadata = None
        app = self.app
        if app is None:
            self._render_status_text()
            return

        mode = getattr(app, "current_mode", getattr(app, "agent_mode", "unknown"))
        model = getattr(app, "current_model", "unknown")
        reasoning = getattr(app, "reasoning_level", "unknown")
        branch = getattr(app, "current_branch", "unknown")
        workspace = "unknown"
        try:
            executor = getattr(app, "executor", None)
            if executor is not None:
                ws = getattr(executor, "workspace_dir", None)
                if ws is not None:
                    workspace = str(ws)
        except Exception:
            pass

        self.update_status(mode, model, reasoning, workspace, branch)

    def _get_display_workspace(self, workspace: str) -> str:
        if workspace == "unknown":
            return workspace
        try:
            # Normalize backslashes immediately for consistent behavior across platforms
            norm_workspace = workspace.replace("\\", "/")
            # Use module-level Path for construction so it can be monkeypatched by tests
            path = Path(norm_workspace)
            # Use pathlib.Path.home() for home detection so it works even if Path
            # is patched to a non-class
            home = pathlib.Path.home()

            if str(path) == str(home) or path == home:
                return "~"

            try:
                # is_relative_to handles path comparison without filesystem access
                if path.is_relative_to(home):
                    rel = path.relative_to(home).as_posix()
                    return f"~/{rel}"
            except (ValueError, TypeError):
                pass

            return path.as_posix()
        except Exception:
            pass
        return workspace.replace("\\", "/")

    def _render_status_text(self) -> None:
        # Compact format: {mode} • {model} ({reasoning}) • {workspace} • {branch}
        workspace_display = self._get_display_workspace(self._workspace)
        parts = [
            self._mode,
            f"{self._model} ({self._reasoning})",
            workspace_display,
            self._branch,
        ]

        if self._turn_cost is not None and self._turn_cost > 0.000001:
            parts.append(f"${self._turn_cost:.3f} turn")
        if self._session_cost is not None and self._session_cost > 0.000001:
            parts.append(f"${self._session_cost:.3f} session")

        text = self.SEPARATOR.join(parts)

        if self._fragment_description is not None and self._fragment_size is not None:
            size_text = format_token_count(self._fragment_size)
            # Label-free fragment: {description} ({tokens} tokens)
            text = f"{self._fragment_description} ({size_text} tokens)"

        self._set_status_metadata(text)

    def _set_status_metadata(self, text: str) -> None:
        metadata = self._metadata
        if metadata is None:
            return
        try:
            metadata.update(text)
        except Exception:
            pass

    def update_status(
        self,
        mode: Optional[str] = None,
        model: Optional[str] = None,
        reasoning: Optional[str] = None,
        workspace: Optional[str] = None,
        branch: Optional[str] = None,
        turn_cost: Optional[float] = None,
        session_cost: Optional[float] = None,
    ) -> None:
        """Update the metadata text segment. Only non-None values are updated."""
        if mode is not None:
            self._mode = str(mode)
        if model is not None:
            self._model = str(model)
        if reasoning is not None:
            self._reasoning = str(reasoning)
        if workspace is not None:
            self._workspace = str(workspace)
        if branch is not None:
            self._branch = str(branch)
        if turn_cost is not None:
            self._turn_cost = turn_cost
        if session_cost is not None:
            self._session_cost = session_cost
        self._render_status_text()

    def set_fragment_info(self, description: Optional[str], size: Optional[int]) -> None:
        description = (description or "").strip()
        self._fragment_description = description or None
        self._fragment_size = size if size is not None and size >= 0 else None
        self._render_status_text()

    def clear_fragment_info(self) -> None:
        self._fragment_description = None
        self._fragment_size = None
        self._render_status_text()

    def set_job_running(self, running: bool) -> None:
        """
        Update internal job state.
        Note: The elapsed timer was moved to ChatPanel help row; this is a no-op for StatusLine.
        """
        pass



from pathlib import Path
from unittest.mock import MagicMock

import pytest
from textual.app import App, ComposeResult

from brokk_code.widgets.status_line import StatusLine


def test_status_line_rendering_compact_home_abbreviation(monkeypatch):
    # Mock home directory
    fake_home = Path("/home/user")
    monkeypatch.setattr(Path, "home", lambda: fake_home)

    # Mock Path to return our controlled paths so is_relative_to works without resolve()
    original_path = Path

    def mock_path(p):
        return original_path(p)

    monkeypatch.setattr("brokk_code.widgets.status_line.Path", mock_path)

    status = StatusLine()
    mock_metadata = MagicMock()
    status._metadata = mock_metadata

    # Case 1: Exactly home
    status.update_status(
        mode="LUTZ",
        model="gpt-4",
        reasoning="high",
        workspace="/home/user",
        branch="main",
    )
    # Expected format: {mode} • {model} ({reasoning}) • {workspace} • {branch}
    expected_home = "LUTZ • gpt-4 (high) • ~ • main"
    mock_metadata.update.assert_called_with(expected_home)

    # Case 2: Under home (verify partial update preserves previous values)
    status.update_status(
        workspace="/home/user/projects/brokk",
    )
    expected_sub = "LUTZ • gpt-4 (high) • ~/projects/brokk • main"
    mock_metadata.update.assert_called_with(expected_sub)

    # Case 3: Explicit update to "unknown" (verify explicit values overwrite previous values)
    status.update_status(
        mode="unknown",
        branch="unknown",
    )
    expected_unknown = "unknown • gpt-4 (high) • ~/projects/brokk • unknown"
    mock_metadata.update.assert_called_with(expected_unknown)

    # Case 4: Explicit update to empty string
    status.update_status(
        mode="",
    )
    expected_empty = " • gpt-4 (high) • ~/projects/brokk • unknown"
    mock_metadata.update.assert_called_with(expected_empty)


def test_status_line_rendering_compact_no_abbreviation(monkeypatch):
    fake_home = Path("/home/user")
    monkeypatch.setattr(Path, "home", lambda: fake_home)

    status = StatusLine()
    mock_metadata = MagicMock()
    status._metadata = mock_metadata

    # Workspace outside home should remain full path
    status.update_status(
        mode="LUTZ",
        model="gpt-4",
        reasoning="high",
        workspace="/var/www/project",
        branch="main",
    )

    # Format: {mode} • {model} ({reasoning}) • {workspace} • {branch}
    expected = "LUTZ • gpt-4 (high) • /var/www/project • main"
    mock_metadata.update.assert_called_with(expected)


def test_status_line_fragment_rendering():
    status = StatusLine()
    mock_metadata = MagicMock()
    status._metadata = mock_metadata

    status.set_fragment_info("my-file.py", 1234)

    # Expected: {description} ({tokens} tokens)
    from brokk_code.token_format import format_token_count

    size_text = format_token_count(1234)
    expected = f"my-file.py ({size_text} tokens)"
    mock_metadata.update.assert_called_with(expected)


def test_status_line_rendering_windows_path_normalization(monkeypatch):
    # Mock a Windows home directory
    fake_home = Path("C:/Users/user")
    monkeypatch.setattr(Path, "home", lambda: fake_home)

    status = StatusLine()
    mock_metadata = MagicMock()
    status._metadata = mock_metadata

    # Case 1: Windows path outside home with backslashes
    status.update_status(
        mode="LUTZ",
        model="gpt-4",
        reasoning="high",
        workspace="D:\\projects\\external",
        branch="main",
    )
    # Note: Path("D:\\projects\\external").as_posix() or the fallback replace("\\", "/")
    # should result in forward slashes.
    expected_external = "LUTZ • gpt-4 (high) • D:/projects/external • main"
    mock_metadata.update.assert_called_with(expected_external)

    # Case 2: Windows path inside home with backslashes
    # We use forward slashes for the comparison in mock_path logic or let Path handle it
    status.update_status(
        workspace="C:\\Users\\user\\projects\\brokk",
    )
    # Expected: Home abbreviation + forward slashes
    expected_home_sub = "LUTZ • gpt-4 (high) • ~/projects/brokk • main"
    mock_metadata.update.assert_called_with(expected_home_sub)


class StatusLineTestApp(App):
    def compose(self) -> ComposeResult:
        yield StatusLine(id="status-line")


@pytest.mark.asyncio
async def test_status_line_no_timer_regression():
    """
    Ensure StatusLine does not contain legacy timer or progress widgets.
    The timer logic was moved to ChatPanel.
    """
    app = StatusLineTestApp()
    async with app.run_test() as pilot:
        status_line = app.query_one("#status-line", StatusLine)

        # 1. Assert legacy timer/progress widgets are absent
        legacy_selectors = ["#status-timer", "#status-timer-wrap", "#status-progress"]
        for selector in legacy_selectors:
            matches = status_line.query(selector)
            assert len(matches) == 0, f"Found legacy widget {selector} in StatusLine"

        # 2. Verify set_job_running is a safe no-op
        # This ensures that even if called (e.g. from ChatPanel), it doesn't crash
        status_line.set_job_running(True)
        status_line.set_job_running(False)
        await pilot.pause()


def test_status_line_cost_rendering():
    status = StatusLine()
    mock_metadata = MagicMock()
    status._metadata = mock_metadata

    # Case 1: No cost (current behavior)
    status.update_status(
        mode="LUTZ",
        model="gpt-4",
        reasoning="high",
        workspace="/work",
        branch="main",
    )
    expected_no_cost = "LUTZ • gpt-4 (high) • /work • main"
    mock_metadata.update.assert_called_with(expected_no_cost)

    # Case 2: Only turn cost
    status.update_status(turn_cost=0.0123)
    expected_turn_only = "LUTZ • gpt-4 (high) • /work • main • $0.012 turn"
    mock_metadata.update.assert_called_with(expected_turn_only)

    # Case 3: Both turn and session cost
    status.update_status(session_cost=0.4567)
    expected_both = "LUTZ • gpt-4 (high) • /work • main • $0.012 turn • $0.457 session"
    mock_metadata.update.assert_called_with(expected_both)

    # Case 4: Fragment hover takes precedence over cost
    status.set_fragment_info("some-file.txt", 500)
    from brokk_code.token_format import format_token_count

    size_text = format_token_count(500)
    expected_fragment = f"some-file.txt ({size_text} tokens)"
    mock_metadata.update.assert_called_with(expected_fragment)

    # Case 5: Clearing fragment restores cost display
    status.clear_fragment_info()
    mock_metadata.update.assert_called_with(expected_both)



"""Tests for the 'Run selected task' execution flow."""

from pathlib import Path
from typing import Any, AsyncIterator, Dict, List, Optional
from unittest.mock import MagicMock

import pytest

from brokk_code.app import BrokkApp
from brokk_code.executor import ExecutorManager
from brokk_code.widgets.chat_panel import ChatPanel
from brokk_code.widgets.tasklist_panel import TaskListPanel


class StubExecutor(ExecutorManager):
    """Stub executor for testing task execution flow."""

    def __init__(self, workspace_dir: Path):
        super().__init__(workspace_dir=workspace_dir)
        self.submit_calls: List[Dict[str, Any]] = []
        self.set_tasklist_calls: List[Dict[str, Any]] = []
        self._tasklist_data: Dict[str, Any] = {"bigPicture": None, "tasks": []}

    async def start(self) -> None:
        pass

    async def stop(self) -> None:
        pass

    async def create_session(self, name: str = "TUI Session") -> str:
        self.session_id = "session-1"
        return self.session_id

    async def wait_ready(self, timeout: float = 30.0) -> bool:
        return True

    def check_alive(self) -> bool:
        return True

    async def get_health_live(self) -> Dict[str, Any]:
        return {"version": "test", "protocolVersion": "1", "execId": "test-id"}

    async def get_context(self) -> Dict[str, Any]:
        return {"fragments": [], "usedTokens": 0, "maxTokens": 100000, "branch": "main"}

    async def get_tasklist(self) -> Dict[str, Any]:
        return self._tasklist_data

    async def set_tasklist(self, tasklist_data: Dict[str, Any]) -> Dict[str, Any]:
        self.set_tasklist_calls.append(tasklist_data)
        self._tasklist_data = tasklist_data
        return tasklist_data

    async def submit_job(
        self,
        task_input: str,
        planner_model: str,
        code_model: Optional[str] = None,
        reasoning_level: Optional[str] = None,
        reasoning_level_code: Optional[str] = None,
        mode: str = "LUTZ",
        tags: Optional[Dict[str, str]] = None,
        session_id: Optional[str] = None,
        auto_commit: bool = True,
        skip_verification: Optional[bool] = None,
        max_issue_fix_attempts: Optional[int] = None,
    ) -> str:
        self.submit_calls.append(
            {
                "task_input": task_input,
                "planner_model": planner_model,
                "code_model": code_model,
                "reasoning_level": reasoning_level,
                "reasoning_level_code": reasoning_level_code,
                "mode": mode,
                "auto_commit": auto_commit,
            }
        )
        return "job-1"

    async def stream_events(self, job_id: str) -> AsyncIterator[Dict[str, Any]]:
        yield {"type": "LLM_TOKEN", "data": {"token": "done", "isTerminal": True}}

    async def cancel_job(self, job_id: str) -> None:
        pass


def _close_coro(coro):
    """Helper to immediately close background coroutines started by run_worker."""
    coro.close()


@pytest.mark.asyncio
async def test_run_selected_task_submits_job_with_task_text(tmp_path: Path) -> None:
    """Verify that running a task submits a job with the task's text."""
    executor = StubExecutor(tmp_path)
    executor._tasklist_data = {
        "bigPicture": "Test goal",
        "tasks": [
            {"id": "task-1", "title": "First Task", "text": "Do the first thing", "done": False},
            {"id": "task-2", "title": "Second Task", "text": "Do the second thing", "done": False},
        ],
    }

    app = BrokkApp(executor=executor)
    app._executor_ready = True
    app.current_model = "test-model"
    app.code_model = "test-code-model"
    app.reasoning_level = "low"
    app.reasoning_level_code = "disable"
    app.auto_commit = True

    # Simulate the selected task
    task = {"id": "task-1", "title": "First Task", "text": "Do the first thing", "done": False}

    await app._run_selected_task(task)

    # Verify submit was called with correct parameters
    assert len(executor.submit_calls) == 1
    call = executor.submit_calls[0]
    assert call["task_input"] == "Do the first thing"
    assert call["planner_model"] == "test-model"
    assert call["code_model"] == "test-code-model"
    assert call["mode"] == "CODE"
    assert call["auto_commit"] is True


@pytest.mark.asyncio
async def test_run_selected_task_falls_back_to_title_when_text_empty(tmp_path: Path) -> None:
    """Verify that running a task uses title when text is empty."""
    executor = StubExecutor(tmp_path)

    app = BrokkApp(executor=executor)
    app._executor_ready = True
    app.current_model = "test-model"

    # Task with empty text
    task = {"id": "task-1", "title": "Use this title", "text": "", "done": False}

    await app._run_selected_task(task)

    assert len(executor.submit_calls) == 1
    assert executor.submit_calls[0]["task_input"] == "Use this title"


@pytest.mark.asyncio
async def test_run_selected_task_marks_task_done_on_success(tmp_path: Path) -> None:
    """Verify that the task is marked done after successful execution."""
    executor = StubExecutor(tmp_path)
    executor._tasklist_data = {
        "bigPicture": "Test goal",
        "tasks": [
            {"id": "task-1", "title": "First Task", "text": "Do the first thing", "done": False},
            {"id": "task-2", "title": "Second Task", "text": "Do the second thing", "done": False},
        ],
    }

    app = BrokkApp(executor=executor)
    app._executor_ready = True
    app.current_model = "test-model"

    task = {"id": "task-1", "title": "First Task", "text": "Do the first thing", "done": False}

    await app._run_selected_task(task)

    # Verify set_tasklist was called
    assert len(executor.set_tasklist_calls) == 1
    saved_data = executor.set_tasklist_calls[0]

    # Find the task and verify it's marked done
    tasks = saved_data.get("tasks", [])
    task_1 = next((t for t in tasks if t.get("id") == "task-1"), None)
    task_2 = next((t for t in tasks if t.get("id") == "task-2"), None)

    assert task_1 is not None
    assert task_1["done"] is True

    # Other tasks should remain unchanged
    assert task_2 is not None
    assert task_2["done"] is False


@pytest.mark.asyncio
async def test_run_selected_task_does_not_mark_done_on_failure(tmp_path: Path) -> None:
    """Verify that the task is NOT marked done if execution fails."""

    class FailingExecutor(StubExecutor):
        async def stream_events(self, job_id: str) -> AsyncIterator[Dict[str, Any]]:
            raise RuntimeError("Simulated failure")
            yield  # Make this a generator

    executor = FailingExecutor(tmp_path)
    executor._tasklist_data = {
        "bigPicture": "Test goal",
        "tasks": [
            {"id": "task-1", "title": "First Task", "text": "Do the first thing", "done": False},
        ],
    }

    app = BrokkApp(executor=executor)
    app._executor_ready = True
    app.current_model = "test-model"

    task = {"id": "task-1", "title": "First Task", "text": "Do the first thing", "done": False}

    await app._run_selected_task(task)

    # set_tasklist should NOT be called since the job failed
    assert len(executor.set_tasklist_calls) == 0


@pytest.mark.parametrize(
    "executor_ready, job_in_progress, selected_task, expected_message_fragment",
    [
        pytest.param(True, False, None, "No task selected", id="no_task_selected"),
        pytest.param(
            True,
            True,
            {"id": "1", "title": "T", "text": "T", "done": False},
            "already in progress",
            id="job_in_progress",
        ),
        pytest.param(
            False,
            False,
            {"id": "1", "title": "T", "text": "T", "done": False},
            "not ready",
            id="executor_not_ready",
        ),
        pytest.param(
            True,
            False,
            {"id": "1", "title": "T", "text": "T", "done": True},
            "already done",
            id="task_already_done",
        ),
    ],
)
def test_action_task_run_guard_conditions(
    executor_ready: bool,
    job_in_progress: bool,
    selected_task,
    expected_message_fragment: str,
) -> None:
    """Verify that action_task_run rejects invalid states with appropriate messages."""
    app = BrokkApp(executor=MagicMock())
    app._executor_ready = executor_ready
    app.job_in_progress = job_in_progress

    mock_chat = MagicMock(spec=ChatPanel)
    mock_panel = MagicMock(spec=TaskListPanel)
    mock_panel.selected_task.return_value = selected_task

    def query_one(target, *args, **kwargs):
        if target is ChatPanel:
            return mock_chat
        if target == "#side-tasklist":
            return mock_panel
        raise AssertionError(f"Unexpected query target: {target}")

    app.query_one = MagicMock(side_effect=query_one)
    app.run_worker = MagicMock(side_effect=_close_coro)

    app.action_task_run()

    mock_chat.add_system_message.assert_called_once()
    assert expected_message_fragment in mock_chat.add_system_message.call_args.args[0]
    app.run_worker.assert_not_called()


@pytest.mark.asyncio
async def test_run_all_incomplete_tasks_executes_in_order(tmp_path: Path) -> None:
    """Verify that run_all executes tasks in order and marks each done."""
    executor = StubExecutor(tmp_path)
    executor._tasklist_data = {
        "bigPicture": "Test goal",
        "tasks": [
            {"id": "task-1", "title": "First", "text": "Do first", "done": False},
            {"id": "task-2", "title": "Second", "text": "Do second", "done": False},
            {"id": "task-3", "title": "Third", "text": "Do third", "done": True},  # Already done
        ],
    }

    app = BrokkApp(executor=executor)
    app._executor_ready = True
    app.current_model = "test-model"

    await app._run_all_incomplete_tasks()

    # Should have submitted 2 jobs (skipping the done task)
    assert len(executor.submit_calls) == 2
    assert executor.submit_calls[0]["task_input"] == "Do first"
    assert executor.submit_calls[1]["task_input"] == "Do second"

    # Both incomplete tasks should now be marked done
    final_data = executor._tasklist_data
    tasks = final_data.get("tasks", [])
    task_1 = next((t for t in tasks if t.get("id") == "task-1"), None)
    task_2 = next((t for t in tasks if t.get("id") == "task-2"), None)
    task_3 = next((t for t in tasks if t.get("id") == "task-3"), None)

    assert task_1 is not None and task_1["done"] is True
    assert task_2 is not None and task_2["done"] is True
    assert task_3 is not None and task_3["done"] is True  # Was already done


@pytest.mark.asyncio
async def test_run_all_incomplete_tasks_stops_on_failure(tmp_path: Path) -> None:
    """Verify that run_all stops on first failure and leaves remaining tasks unchanged."""

    class FailingOnSecondExecutor(StubExecutor):
        def __init__(self, workspace_dir: Path):
            super().__init__(workspace_dir)
            self._job_count = 0

        async def stream_events(self, job_id: str) -> AsyncIterator[Dict[str, Any]]:
            self._job_count += 1
            if self._job_count == 2:
                raise RuntimeError("Simulated failure on second task")
            yield {"type": "LLM_TOKEN", "data": {"token": "done", "isTerminal": True}}

    executor = FailingOnSecondExecutor(tmp_path)
    executor._tasklist_data = {
        "bigPicture": "Test goal",
        "tasks": [
            {"id": "task-1", "title": "First", "text": "Do first", "done": False},
            {"id": "task-2", "title": "Second", "text": "Do second", "done": False},
            {"id": "task-3", "title": "Third", "text": "Do third", "done": False},
        ],
    }

    app = BrokkApp(executor=executor)
    app._executor_ready = True
    app.current_model = "test-model"

    await app._run_all_incomplete_tasks()

    # Should have submitted 2 jobs (first succeeds, second fails, third never runs)
    assert len(executor.submit_calls) == 2

    # Only the first task should be marked done
    final_data = executor._tasklist_data
    tasks = final_data.get("tasks", [])
    task_1 = next((t for t in tasks if t.get("id") == "task-1"), None)
    task_2 = next((t for t in tasks if t.get("id") == "task-2"), None)
    task_3 = next((t for t in tasks if t.get("id") == "task-3"), None)

    assert task_1 is not None and task_1["done"] is True  # Succeeded
    assert task_2 is not None and task_2["done"] is False  # Failed
    assert task_3 is not None and task_3["done"] is False  # Never ran


@pytest.mark.asyncio
async def test_run_all_incomplete_tasks_persists_after_each_success(tmp_path: Path) -> None:
    """Verify that set_tasklist is called after each successful task."""
    executor = StubExecutor(tmp_path)
    executor._tasklist_data = {
        "bigPicture": "Test goal",
        "tasks": [
            {"id": "task-1", "title": "First", "text": "Do first", "done": False},
            {"id": "task-2", "title": "Second", "text": "Do second", "done": False},
        ],
    }

    app = BrokkApp(executor=executor)
    app._executor_ready = True
    app.current_model = "test-model"

    await app._run_all_incomplete_tasks()

    # set_tasklist should be called twice (once after each task)
    assert len(executor.set_tasklist_calls) == 2

    # First call should have task-1 done, task-2 not done
    first_save = executor.set_tasklist_calls[0]
    first_tasks = first_save.get("tasks", [])
    first_task_1 = next((t for t in first_tasks if t.get("id") == "task-1"), None)
    first_task_2 = next((t for t in first_tasks if t.get("id") == "task-2"), None)
    assert first_task_1 is not None and first_task_1["done"] is True
    assert first_task_2 is not None and first_task_2["done"] is False

    # Second call should have both done
    second_save = executor.set_tasklist_calls[1]
    second_tasks = second_save.get("tasks", [])
    second_task_1 = next((t for t in second_tasks if t.get("id") == "task-1"), None)
    second_task_2 = next((t for t in second_tasks if t.get("id") == "task-2"), None)
    assert second_task_1 is not None and second_task_1["done"] is True
    assert second_task_2 is not None and second_task_2["done"] is True


@pytest.mark.asyncio
async def test_run_all_incomplete_tasks_no_incomplete(tmp_path: Path) -> None:
    """Verify that run_all handles no incomplete tasks gracefully."""
    executor = StubExecutor(tmp_path)
    executor._tasklist_data = {
        "bigPicture": "Test goal",
        "tasks": [
            {"id": "task-1", "title": "First", "text": "Do first", "done": True},
        ],
    }

    app = BrokkApp(executor=executor)
    app._executor_ready = True
    app.current_model = "test-model"

    await app._run_all_incomplete_tasks()

    # No jobs should be submitted
    assert len(executor.submit_calls) == 0


@pytest.mark.asyncio
async def test_run_all_incomplete_tasks_stops_on_cancellation(tmp_path: Path) -> None:
    """Verify that run_all stops when a task is cancelled and does not mark it done."""

    class CancellingExecutor(StubExecutor):
        """Executor that simulates cancellation during the first task."""

        def __init__(self, workspace_dir: Path, app_ref: Optional[Any] = None):
            super().__init__(workspace_dir)
            self._app_ref = app_ref

        def set_app(self, app: Any) -> None:
            self._app_ref = app

        async def stream_events(self, job_id: str) -> AsyncIterator[Dict[str, Any]]:
            # Simulate some progress, then cancellation is triggered
            yield {"type": "LLM_TOKEN", "data": {"token": "working", "isTerminal": False}}
            # Simulate the cancellation flag being set (as if Ctrl+C was pressed)
            if self._app_ref is not None:
                self._app_ref._task_run_cancelled = True
            # Yield terminal event
            yield {"type": "LLM_TOKEN", "data": {"token": "", "isTerminal": True}}

    executor = CancellingExecutor(tmp_path)
    executor._tasklist_data = {
        "bigPicture": "Test goal",
        "tasks": [
            {"id": "task-1", "title": "First", "text": "Do first", "done": False},
            {"id": "task-2", "title": "Second", "text": "Do second", "done": False},
            {"id": "task-3", "title": "Third", "text": "Do third", "done": False},
        ],
    }

    app = BrokkApp(executor=executor)
    executor.set_app(app)
    app._executor_ready = True
    app.current_model = "test-model"

    await app._run_all_incomplete_tasks()

    # Only the first task should have been submitted (cancelled during execution)
    assert len(executor.submit_calls) == 1
    assert executor.submit_calls[0]["task_input"] == "Do first"

    # No tasks should be marked done since the first was cancelled
    final_data = executor._tasklist_data
    tasks = final_data.get("tasks", [])
    task_1 = next((t for t in tasks if t.get("id") == "task-1"), None)
    task_2 = next((t for t in tasks if t.get("id") == "task-2"), None)
    task_3 = next((t for t in tasks if t.get("id") == "task-3"), None)

    assert task_1 is not None and task_1["done"] is False  # Cancelled, not marked done
    assert task_2 is not None and task_2["done"] is False  # Never ran
    assert task_3 is not None and task_3["done"] is False  # Never ran

    # set_tasklist should NOT have been called since no task completed successfully
    assert len(executor.set_tasklist_calls) == 0


@pytest.mark.asyncio
async def test_run_selected_task_does_not_mark_done_on_cancellation(tmp_path: Path) -> None:
    """Verify that a single task is NOT marked done if cancelled."""

    class CancellingExecutor(StubExecutor):
        """Executor that simulates cancellation during task execution."""

        def __init__(self, workspace_dir: Path, app_ref: Optional[Any] = None):
            super().__init__(workspace_dir)
            self._app_ref = app_ref

        def set_app(self, app: Any) -> None:
            self._app_ref = app

        async def stream_events(self, job_id: str) -> AsyncIterator[Dict[str, Any]]:
            yield {"type": "LLM_TOKEN", "data": {"token": "working", "isTerminal": False}}
            # Simulate cancellation
            if self._app_ref is not None:
                self._app_ref._task_run_cancelled = True
            yield {"type": "LLM_TOKEN", "data": {"token": "", "isTerminal": True}}

    executor = CancellingExecutor(tmp_path)
    executor._tasklist_data = {
        "bigPicture": "Test goal",
        "tasks": [
            {"id": "task-1", "title": "First Task", "text": "Do the first thing", "done": False},
        ],
    }

    app = BrokkApp(executor=executor)
    executor.set_app(app)
    app._executor_ready = True
    app.current_model = "test-model"

    task = {"id": "task-1", "title": "First Task", "text": "Do the first thing", "done": False}

    await app._run_selected_task(task)

    # Job was submitted
    assert len(executor.submit_calls) == 1

    # Task should NOT be marked done since it was cancelled
    assert len(executor.set_tasklist_calls) == 0
